import pytest

from vyper import compiler
from vyper.exceptions import (
    ArgumentException,
    InterfaceViolation,
    InvalidReference,
    InvalidType,
    ModuleNotFound,
    NamespaceCollision,
    StructureException,
    SyntaxException,
    TypeMismatch,
    UnknownAttribute,
)

fail_list = [
    (
        """
from ethereum.ercs import IERC20
a: public(IERC20)
@external
def test():
    b: uint256 = self.a
    """,
        TypeMismatch,
    ),
    (
        """
from ethereum.ercs import IERC20
aba: public(IERC20)
@external
def test():
    self.aba = IERC20
    """,
        InvalidReference,
    ),
    (
        """
from ethereum.ercs import IERC20

a: address(IERC20) # invalid syntax now.
    """,
        SyntaxException,
    ),
    (
        """
from ethereum.ercs import IERC20

@external
def test():
    a: address(IERC20) = empty(address)
    """,
        InvalidType,
    ),
    (
        """
a: address

@external
def test():  # may not call normal address
    assert self.a.random()
    """,
        UnknownAttribute,
    ),
    (
        """
from ethereum.ercs import IERC20
@external
def test(a: address):
    my_address: address = IERC20()
    """,
        ArgumentException,
    ),
    (
        """
from ethereum.ercs import IERC20

implements: IERC20 = 1
    """,
        SyntaxException,
    ),
    (
        """
interface A:
    @external
    def foo(): nonpayable
    """,
        StructureException,
    ),
    (
        """
implements: self.x
    """,
        InvalidType,
    ),
    (
        """
implements: 123
    """,
        StructureException,
    ),
    (
        """
struct Foo:
    a: uint256

implements: Foo
    """,
        StructureException,
    ),
    (
        """
from ethereum.ercs import IERC20

interface A:
    def f(): view

@internal
def foo():
    a: IERC20 = A(empty(address))
    """,
        TypeMismatch,
    ),
    (
        """
interface A:
    def f(a: uint256): view

implements: A

@external
@nonpayable
def f(a: uint256): # visibility is nonpayable instead of view
    pass
    """,
        InterfaceViolation,
    ),
    (
        # exports two Transfer events
        """
from ethereum.ercs import IERC20

implements: IERC20

event Transfer:
    sender: indexed(address)
    receiver: address
    value: uint256

name: public(String[32])
symbol: public(String[32])
decimals: public(uint8)
balanceOf: public(HashMap[address, uint256])
allowance: public(HashMap[address, HashMap[address, uint256]])
totalSupply: public(uint256)

@external
def transfer(_to : address, _value : uint256) -> bool:
    log Transfer(msg.sender, _to, _value)
    return True

@external
def transferFrom(_from : address, _to : address, _value : uint256) -> bool:
    log IERC20.Transfer(_from, _to, _value)
    return True

@external
def approve(_spender : address, _value : uint256) -> bool:
    return True
    """,
        NamespaceCollision,
    ),
    (
        # `payable` decorator not implemented
        """
interface testI:
    def foo() -> uint256: payable

implements: testI

@external
def foo() -> uint256:
    return 0
    """,
        InterfaceViolation,
    ),
    (
        # decorators must be strictly identical
        """
interface Self:
    def protected_view_fn() -> String[100]: nonpayable

implements: Self

@external
@pure
def protected_view_fn() -> String[100]:
    return empty(String[100])
    """,
        InterfaceViolation,
    ),
    (
        # decorators must be strictly identical
        """
interface Self:
    def protected_view_fn() -> String[100]: view

implements: Self

@external
@pure
def protected_view_fn() -> String[100]:
    return empty(String[100])
    """,
        InterfaceViolation,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_interfaces_fail(bad_code):
    with pytest.raises(bad_code[1]):
        compiler.compile_code(bad_code[0])


valid_list = [
    """
from ethereum.ercs import IERC20
b: IERC20
@external
def test(input: address):
    assert staticcall self.b.totalSupply() == staticcall IERC20(input).totalSupply()
    """,
    """
from ethereum.ercs import IERC20

interface Factory:
   def getExchange(token_addr: address) -> address: view

factory: Factory
token: IERC20

@external
def test():
    assert staticcall self.factory.getExchange(self.token.address) == self
    exchange: address = staticcall self.factory.getExchange(self.token.address)
    assert exchange == self.token.address
    assert staticcall self.token.totalSupply() > 0
    """,
    """
interface Foo:
    def foo(): view

@external
def test() -> (bool, Foo):
    x: Foo = Foo(msg.sender)
    return True, x
    """
    """
from ethereum.ercs import IERC20

a: public(IERC20)
    """,
    """
from ethereum.ercs import IERC20

a: public(IERC20)

@external
def test() -> address:
    return self.a.address
    """,
    """
from ethereum.ercs import IERC20

a: public(IERC20)
b: address

@external
def test():
    self.b = self.a.address
    """,
    """
from ethereum.ercs import IERC20

struct aStruct:
   my_address: address

a: public(IERC20)
b: aStruct

@external
def test() -> address:
    self.b.my_address = self.a.address
    return self.b.my_address
    """,
    """
from ethereum.ercs import IERC20
a: public(IERC20)
@external
def test():
    b: address = self.a.address
    """,
    """
interface MyInterface:
    def some_func(): nonpayable

my_interface: MyInterface[3]
idx: uint256

@deploy
def __init__():
    self.my_interface[self.idx] = MyInterface(empty(address))
    """,
    """
interface MyInterface:
    def kick(): payable

kickers: HashMap[address, MyInterface]
    """,
    """
interface Foo:
    def append(a: uint256): payable

@external
def bar(x: address):
    a: Foo = Foo(x)
    extcall a.append(1)
    """,
    """
interface Foo:
    def pop(): payable

@external
def foo(x: address):
    a: Foo = Foo(x)
    extcall a.pop()
    """,
    """
interface ITestInterface:
    def foo() -> uint256: view

implements: ITestInterface

foo: public(constant(uint256)) = 1
    """,
    """
interface ITestInterface:
    def foo() -> uint256: view

implements: ITestInterface

foo: public(immutable(uint256))

@deploy
def __init__(x: uint256):
    foo = x
    """,
    # no namespace collision of interface after storage variable
    """
a: constant(uint256) = 1

interface A:
    def f(a: uint128): view
    """,
    # no namespace collision of storage variable after interface
    """
interface A:
    def f(a: uint256): view

a: constant(uint128) = 1
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_interfaces_success(good_code):
    assert compiler.compile_code(good_code) is not None


def test_imports_and_implements_within_interface(make_input_bundle):
    interface_code = """
@external
def foobar():
    ...
"""

    input_bundle = make_input_bundle({"foo.vyi": interface_code})

    code = """
import foo as Foo

implements: Foo

@external
def foobar():
    pass
"""

    assert compiler.compile_code(code, input_bundle=input_bundle) is not None


def test_builtins_not_found():
    code = """
from vyper.interfaces import foobar
    """
    with pytest.raises(ModuleNotFound) as e:
        compiler.compile_code(code)

    assert e.value._message == "vyper.interfaces.foobar"
    assert e.value._hint == "try renaming `vyper.interfaces` to `ethereum.ercs`"


@pytest.mark.parametrize("erc", ("ERC20", "ERC721", "ERC4626"))
def test_builtins_not_found2(erc):
    code = f"""
from ethereum.ercs import {erc}
    """
    with pytest.raises(ModuleNotFound) as e:
        compiler.compile_code(code)
    assert e.value._message == f"ethereum.ercs.{erc}"
    assert e.value._hint == f"try renaming `{erc}` to `I{erc}`"



import pytest
from hypothesis import assume, given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold

st_uint256 = st.integers(min_value=0, max_value=2**256 - 1)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(a=st_uint256, b=st_uint256, c=st_uint256)
@pytest.mark.parametrize("fn_name", ["uint256_addmod", "uint256_mulmod"])
def test_modmath(get_contract, a, b, c, fn_name):
    assume(c > 0)

    source = f"""
@external
def foo(a: uint256, b: uint256, c: uint256) -> uint256:
    return {fn_name}(a, b, c)
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{fn_name}({a}, {b}, {c})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(a, b, c) == new_node.value



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import InvalidLiteral, InvalidOperation, TypeMismatch
from vyper.utils import unsigned_to_signed

code = """
@external
def _bitwise_and(x: uint256, y: uint256) -> uint256:
    return x & y

@external
def _bitwise_or(x: uint256, y: uint256) -> uint256:
    return x | y

@external
def _bitwise_xor(x: uint256, y: uint256) -> uint256:
    return x ^ y

@external
def _bitwise_not(x: uint256) -> uint256:
    return ~x

@external
def _shl(x: uint256, y: uint256) -> uint256:
    return x << y

@external
def _shr(x: uint256, y: uint256) -> uint256:
    return x >> y
    """


def test_bitwise_opcodes():
    opcodes = compile_code(code, output_formats=["opcodes"])["opcodes"]
    assert "SHL" in opcodes
    assert "SHR" in opcodes


def test_test_bitwise(get_contract):
    c = get_contract(code)
    x = 126416208461208640982146408124
    y = 7128468721412412459
    assert c._bitwise_and(x, y) == (x & y)
    assert c._bitwise_or(x, y) == (x | y)
    assert c._bitwise_xor(x, y) == (x ^ y)
    assert c._bitwise_not(x) == 2**256 - 1 - x

    for t in (x, y):
        for s in (0, 1, 3, 255, 256):
            assert c._shr(t, s) == t >> s
            assert c._shl(t, s) == (t << s) % (2**256)


def test_signed_shift(get_contract):
    code = """
@external
def _sar(x: int256, y: uint256) -> int256:
    return x >> y

@external
def _shl(x: int256, y: uint256) -> int256:
    return x << y
    """
    c = get_contract(code)
    x = 126416208461208640982146408124
    y = 7128468721412412459
    cases = [x, y, -x, -y]

    for t in cases:
        for s in (0, 1, 3, 255, 256):
            assert c._sar(t, s) == t >> s
            assert c._shl(t, s) == unsigned_to_signed((t << s) % (2**256), 256)


def test_precedence(get_contract):
    code = """
@external
def foo(a: uint256, b: uint256, c: uint256) -> (uint256, uint256):
    return (a | b & c, (a | b) & c)

@external
def bar(a: uint256, b: uint256, c: uint256) -> (uint256, uint256):
    return (a | ~b & c, (a | ~b) & c)

@external
def baz(a: uint256, b: uint256, c: uint256) -> (uint256, uint256):
    return (a + 8 | ~b & c * 2, (a  + 8 | ~b) & c * 2)
    """
    c = get_contract(code)
    assert tuple(c.foo(1, 6, 14)) == (1 | 6 & 14, (1 | 6) & 14) == (7, 6)
    assert tuple(c.bar(1, 6, 14)) == (1 | ~6 & 14, (1 | ~6) & 14) == (9, 8)
    assert tuple(c.baz(1, 6, 14)) == (1 + 8 | ~6 & 14 * 2, (1 + 8 | ~6) & 14 * 2) == (25, 24)


def test_literals(get_contract):
    code = """
@external
def _shr(x: uint256) -> uint256:
    return x >> 3

@external
def _shl(x: uint256) -> uint256:
    return x << 3
    """

    c = get_contract(code)
    assert c._shr(80) == 10
    assert c._shl(80) == 640


fail_list = [
    (
        # cannot shift non-uint256/int256 argument
        """
@external
def foo(x: uint8, y: uint8) -> uint8:
    return x << y
    """,
        InvalidOperation,
    ),
    (
        # cannot shift non-uint256/int256 argument
        """
@external
def foo(x: int8, y: uint8) -> int8:
    return x << y
    """,
        InvalidOperation,
    ),
    (
        # cannot shift by non-uint bits
        """
@external
def foo(x: uint256, y: int128) -> uint256:
    return x << y
    """,
        TypeMismatch,
    ),
    (
        # cannot left shift by more than 256 bits
        """
@external
def foo() -> uint256:
    return 2 << 257
    """,
        InvalidLiteral,
    ),
    (
        # cannot shift by negative amount
        """
@external
def foo() -> uint256:
    return 2 << -1
    """,
        InvalidLiteral,
    ),
    (
        # cannot shift by negative amount
        """
@external
def foo() -> uint256:
    return 2 << -1
    """,
        InvalidLiteral,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_shift_fail(get_contract, bad_code, exc, assert_compile_failed):
    assert_compile_failed(lambda: get_contract(bad_code), exc)



from pytest import raises

from vyper.exceptions import UndeclaredDefinition


def test_permanent_variables_test(get_contract):
    permanent_variables_test = """
struct Var:
    a: int128
    b: int128
var: Var

@deploy
def __init__(a: int128, b: int128):
    self.var.a = a
    self.var.b = b

@external
def returnMoose() -> int128:
    return self.var.a * 10 + self.var.b
    """

    c = get_contract(permanent_variables_test, *[5, 7])
    assert c.returnMoose() == 57
    print("Passed init argument and variable member test")


def test_missing_global(get_contract):
    code = """
@external
def a() -> int128:
    return self.b
    """

    with raises(UndeclaredDefinition):
        get_contract(code)



import pytest

from tests.utils import ZERO_ADDRESS

MAX_BIDS = 128

BIDDING_TIME = 150
REVEAL_TIME = 50
TEST_INCREMENT = 1


@pytest.fixture(scope="module")
def auction_contract(env, get_contract):
    with open("examples/auctions/blind_auction.vy") as f:
        contract_code = f.read()

    for acc in env.accounts[1:4]:
        env.set_balance(acc, 10**20)

    return get_contract(contract_code, *[env.deployer, BIDDING_TIME, REVEAL_TIME])


def test_initial_state(env, auction_contract):
    # Check beneficiary is correct
    assert auction_contract.beneficiary() == env.deployer
    # Check that bidding end time is correct
    assert auction_contract.biddingEnd() == env.timestamp + BIDDING_TIME
    # Check that the reveal end time is correct
    assert auction_contract.revealEnd() == auction_contract.biddingEnd() + REVEAL_TIME
    # Check auction has not ended
    assert auction_contract.ended() is False
    # Check highest bid is 0
    assert auction_contract.highestBid() == 0
    # Check highest bidder is empty
    assert auction_contract.highestBidder() == ZERO_ADDRESS


def test_late_bid(env, auction_contract, tx_failed, keccak):
    k1 = env.accounts[1]

    # Move time forward past bidding end
    env.timestamp += BIDDING_TIME + TEST_INCREMENT

    # Try to bid after bidding has ended
    with tx_failed():
        auction_contract.bid(
            keccak(
                b"".join(
                    [
                        (200).to_bytes(32, byteorder="big"),
                        (0).to_bytes(32, byteorder="big"),
                        (8675309).to_bytes(32, byteorder="big"),
                    ]
                )
            ),
            value=200,
            sender=k1,
        )


def test_too_many_bids(env, auction_contract, tx_failed, keccak):
    k1 = env.accounts[1]

    # First 128 bids should be able to be placed successfully
    for i in range(MAX_BIDS):
        auction_contract.bid(
            keccak(
                b"".join(
                    [
                        (i).to_bytes(32, byteorder="big"),
                        (1).to_bytes(32, byteorder="big"),
                        (8675309).to_bytes(32, byteorder="big"),
                    ]
                )
            ),
            value=i,
            sender=k1,
        )

    # 129th bid should fail
    with tx_failed():
        auction_contract.bid(
            keccak(
                b"".join(
                    [
                        (128).to_bytes(32, byteorder="big"),
                        (0).to_bytes(32, byteorder="big"),
                        (8675309).to_bytes(32, byteorder="big"),
                    ]
                )
            ),
            value=128,
            sender=k1,
        )


def test_early_reval(env, auction_contract, tx_failed, keccak):
    k1 = env.accounts[1]

    # k1 places 1 real bid
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (100).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (8675309).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=100,
        sender=k1,
    )

    # Move time slightly forward (still before bidding has ended)
    env.timestamp += TEST_INCREMENT

    # Try to reveal early
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _numBids = 1
    _values[0] = 100
    _fakes[0] = False
    _secrets[0] = (8675309).to_bytes(32, byteorder="big")
    with tx_failed():
        auction_contract.reveal(_numBids, _values, _fakes, _secrets, value=0, sender=k1)

    # Check highest bidder is still empty
    assert auction_contract.highestBidder() == ZERO_ADDRESS
    # Check highest bid is still 0
    assert auction_contract.highestBid() == 0


def test_late_reveal(env, auction_contract, tx_failed, keccak):
    k1 = env.accounts[1]

    # k1 places 1 real bid
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (100).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (8675309).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=100,
        sender=k1,
    )

    # Move time forward past bidding _and_ reveal time
    env.timestamp += BIDDING_TIME + REVEAL_TIME + TEST_INCREMENT

    # Try to reveal late
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _numBids = 1
    _values[0] = 100
    _fakes[0] = False
    _secrets[0] = (8675309).to_bytes(32, byteorder="big")
    with tx_failed():
        auction_contract.reveal(_numBids, _values, _fakes, _secrets, value=0, sender=k1)

    # Check highest bidder is still empty
    assert auction_contract.highestBidder() == ZERO_ADDRESS
    # Check highest bid is still 0
    assert auction_contract.highestBid() == 0


def test_early_end(env, auction_contract, tx_failed):
    k0 = env.deployer

    # Should not be able to end auction before reveal time has ended
    with tx_failed():
        auction_contract.auctionEnd(value=0, sender=k0)


def test_double_end(env, auction_contract, tx_failed):
    k0 = env.deployer

    # Move time forward past bidding and reveal end
    env.timestamp += BIDDING_TIME + REVEAL_TIME + TEST_INCREMENT

    # First auction end should succeed
    auction_contract.auctionEnd(value=0, sender=k0)

    # Should not be able to end auction twice
    with tx_failed():
        auction_contract.auctionEnd(value=0, sender=k0)


def test_blind_auction(env, auction_contract, keccak):
    k0, k1, k2, k3 = env.accounts[0:4]
    for acc in env.accounts[1:4]:
        env.set_balance(acc, 10**20)

    ###################################################################
    #                         Place bids                              #
    ###################################################################

    # k1 places 1 real bid
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (100).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (8675309).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=100,
        sender=k1,
    )

    # k2 places 1 real bid (highest) and 2 fake
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (150).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (1234567).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=150,
        sender=k2,
    )
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (200).to_bytes(32, byteorder="big"),
                    (0).to_bytes(32, byteorder="big"),
                    (1234567).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=250,
        sender=k2,
    )
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (300).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (1234567).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=300,
        sender=k2,
    )

    # k3 places 2 fake bids
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (175).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (9876543).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=175,
        sender=k3,
    )
    auction_contract.bid(
        keccak(
            b"".join(
                [
                    (275).to_bytes(32, byteorder="big"),
                    (1).to_bytes(32, byteorder="big"),
                    (9876543).to_bytes(32, byteorder="big"),
                ]
            )
        ),
        value=275,
        sender=k3,
    )

    ###################################################################
    #                          Reveal bids.                           #
    ###################################################################

    # Move time forward past bidding end (still within reveal end)
    env.timestamp += BIDDING_TIME + TEST_INCREMENT

    # Reveal k1 bids
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _numBids = 1
    _values[0] = 100
    _fakes[0] = False
    _secrets[0] = (8675309).to_bytes(32, byteorder="big")
    auction_contract.reveal(_numBids, _values, _fakes, _secrets, value=0, sender=k1)

    #: Check that highest bidder and highest bid have updated
    assert auction_contract.highestBid() == 100
    assert auction_contract.highestBidder() == k1

    # Reveal k2 bids
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _values[0] = 150
    _fakes[0] = True
    _secrets[0] = (1234567).to_bytes(32, byteorder="big")
    _values[1] = 200
    _fakes[1] = False
    _secrets[1] = (1234567).to_bytes(32, byteorder="big")
    _values[2] = 300
    _fakes[2] = True
    _secrets[2] = (1234567).to_bytes(32, byteorder="big")
    balance_before_reveal = env.get_balance(k2)
    auction_contract.reveal(3, _values, _fakes, _secrets, value=0, sender=k2)
    balance_after_reveal = env.get_balance(k2)

    #: Check that highest bidder and highest bid have updated
    assert auction_contract.highestBid() == 200
    assert auction_contract.highestBidder() == k2

    # Check for refund from false bids (300 + 150) and partial deposit (50)
    assert balance_after_reveal == (balance_before_reveal + 500)

    # Reveal k3 bids
    _values = [0] * MAX_BIDS  # Initialized with 128 default values
    _fakes = [False] * MAX_BIDS  # Initialized with 128 default values
    _secrets = [b"\x00" * 32] * MAX_BIDS  # Initialized with 128 default values
    _values[0] = 175
    _fakes[0] = True
    _secrets[0] = (9876543).to_bytes(32, byteorder="big")
    _values[1] = 275
    _fakes[1] = True
    _secrets[1] = (9876543).to_bytes(32, byteorder="big")
    balance_before_reveal = env.get_balance(k3)
    auction_contract.reveal(2, _values, _fakes, _secrets, value=0, sender=k3)
    balance_after_reveal = env.get_balance(k3)

    #: Check that highest bidder and highest bid have NOT updated
    assert auction_contract.highestBidder() == k2
    assert auction_contract.highestBid() == 200

    # Check for refund from false bids (175 + 275 = 450)
    assert balance_after_reveal == (balance_before_reveal + 450)

    ###################################################################
    #                          End auction.                           #
    ###################################################################

    # Move time forward past bidding and reveal end
    env.timestamp += REVEAL_TIME

    # End the auction
    balance_before_end = env.get_balance(k0)
    auction_contract.auctionEnd(value=0, sender=k0)
    balance_after_end = env.get_balance(k0)

    # Check that auction indeed ended
    assert auction_contract.ended() is True

    # Check that beneficiary has been paid
    assert balance_after_end == (balance_before_end + 200)

    # Check that k1 is able to withdraw their outbid bid
    balance_before_withdraw = env.get_balance(k1)
    auction_contract.withdraw(value=0, sender=k1)
    balance_after_withdraw = env.get_balance(k1)
    assert balance_after_withdraw == (balance_before_withdraw + 100)



from vyper.utils import hex_to_int


def test_hash_code(get_contract, keccak):
    hash_code = """
@external
def foo(inp: Bytes[100]) -> bytes32:
    return keccak256(inp)

@external
def foob() -> bytes32:
    return keccak256(b"inp")

@external
def bar() -> bytes32:
    return keccak256("inp")
    """

    c = get_contract(hash_code)
    for inp in (b"", b"cow", b"s" * 31, b"\xff" * 32, b"\n" * 33, b"g" * 64, b"h" * 65):
        assert c.foo(inp) == keccak(inp)

    assert c.bar() == keccak(b"inp")
    assert c.foob() == keccak(b"inp")


def test_hash_code2(get_contract):
    hash_code2 = """
@external
def foo(inp: Bytes[100]) -> bool:
    return keccak256(inp) == keccak256("badminton")
    """
    c = get_contract(hash_code2)
    assert c.foo(b"badminto") is False
    assert c.foo(b"badminton") is True


def test_hash_code3(get_contract):
    hash_code3 = """
test: Bytes[100]

@external
def set_test(inp: Bytes[100]):
    self.test = inp

@external
def tryy(inp: Bytes[100]) -> bool:
    return keccak256(inp) == keccak256(self.test)

@external
def tryy_str(inp: String[100]) -> bool:
    return keccak256(inp) == keccak256(self.test)

@external
def trymem(inp: Bytes[100]) -> bool:
    x: Bytes[100] = self.test
    return keccak256(inp) == keccak256(x)

@external
def try32(inp: bytes32) -> bool:
    return keccak256(inp) == keccak256(self.test)

    """
    c = get_contract(hash_code3)
    c.set_test(b"")
    assert c.tryy(b"") is True
    assert c.tryy_str("") is True
    assert c.trymem(b"") is True
    assert c.tryy(b"cow") is False
    c.set_test(b"cow")
    assert c.tryy(b"") is False
    assert c.tryy(b"cow") is True
    assert c.tryy_str("cow") is True
    c.set_test(b"\x35" * 32)
    assert c.tryy(b"\x35" * 32) is True
    assert c.trymem(b"\x35" * 32) is True
    assert c.try32(b"\x35" * 32) is True
    assert c.tryy(b"\x35" * 33) is False
    c.set_test(b"\x35" * 33)
    assert c.tryy(b"\x35" * 32) is False
    assert c.trymem(b"\x35" * 32) is False
    assert c.try32(b"\x35" * 32) is False
    assert c.tryy(b"\x35" * 33) is True

    print("Passed KECCAK256 hash test")


def test_hash_constant_bytes32(get_contract, keccak):
    hex_val = "0x1234567890123456789012345678901234567890123456789012345678901234"
    code = f"""
FOO: constant(bytes32) = {hex_val}
BAR: constant(bytes32) = keccak256(FOO)
@external
def foo() -> bytes32:
    x: bytes32 = BAR
    return x
    """
    c = get_contract(code)
    assert c.foo() == keccak(hex_to_int(hex_val).to_bytes(32, "big"))


def test_hash_constant_string(get_contract, keccak):
    str_val = "0x1234567890123456789012345678901234567890123456789012345678901234"
    code = f"""
FOO: constant(String[66]) = "{str_val}"
BAR: constant(bytes32) = keccak256(FOO)
@external
def foo() -> bytes32:
    x: bytes32 = BAR
    return x
    """
    c = get_contract(code)
    assert c.foo() == keccak(str_val.encode())



def test_nested_struct(get_contract):
    code = """
struct Animal:
    location: address
    fur: String[32]

struct Human:
    location: address
    animal: Animal

@external
def modify_nested_struct(_human: Human) -> Human:
    human: Human = _human

    # do stuff, edit the structs
    # (13 is the length of the result)
    human.animal.fur = slice(concat(human.animal.fur, " is great"), 0, 13)

    return human
    """
    c = get_contract(code)
    addr1 = "0x1234567890123456789012345678901234567890"
    addr2 = "0x1234567890123456789012345678900000000000"
    # assert c.modify_nested_tuple([addr1, 123], [addr2, 456]) == ([addr1, 124], [addr2, 457])
    assert c.modify_nested_struct((addr1, (addr2, "wool"))) == (addr1, (addr2, "wool is great"))


def test_nested_single_struct(get_contract):
    code = """
struct Animal:
    fur: String[32]

struct Human:
    animal: Animal

@external
def modify_nested_single_struct(_human: Human) -> Human:
    human: Human = _human

    # do stuff, edit the structs
    # (13 is the length of the result)
    human.animal.fur = slice(concat(human.animal.fur, " is great"), 0, 13)

    return human
    """
    c = get_contract(code)

    assert c.modify_nested_single_struct((("wool",),)) == (("wool is great",),)



from vyper.compiler import compile_code
from vyper.utils import keccak256


def test_get_extcodehash(get_contract):
    code = """
a: address

@deploy
def __init__():
    self.a = self

@external
def foo(x: address) -> bytes32:
    return x.codehash

@external
def foo2(x: address) -> bytes32:
    b: address = x
    return b.codehash

@external
def foo3() -> bytes32:
    return self.codehash

@external
def foo4() -> bytes32:
    return self.a.codehash
    """
    compiled = compile_code(code, output_formats=["bytecode_runtime"])
    bytecode = bytes.fromhex(compiled["bytecode_runtime"][2:])
    hash_ = keccak256(bytecode)

    c = get_contract(code)

    assert c.foo(c.address) == hash_
    assert not int(c.foo("0xDeaDbeefdEAdbeefdEadbEEFdeadbeEFdEaDbeeF").hex(), 16)

    assert c.foo2(c.address) == hash_
    assert not int(c.foo2("0xDeaDbeefdEAdbeefdEadbEEFdeadbeEFdEaDbeeF").hex(), 16)

    assert c.foo3() == hash_
    assert c.foo4() == hash_



import ast as python_ast
import contextlib
import copy
import decimal
import functools
import math
import operator
import pickle
import sys
import warnings
from typing import Any, Optional, Union

from vyper.ast.metadata import NodeMetadata
from vyper.compiler.settings import VYPER_ERROR_CONTEXT_LINES, VYPER_ERROR_LINE_NUMBERS
from vyper.exceptions import (
    ArgumentException,
    CompilerPanic,
    InvalidLiteral,
    InvalidOperation,
    OverflowException,
    StructureException,
    SyntaxException,
    TypeMismatch,
    UnfoldableNode,
    VariableDeclarationException,
    VyperException,
    ZeroDivisionException,
)
from vyper.utils import (
    MAX_DECIMAL_PLACES,
    SizeLimits,
    annotate_source_code,
    evm_div,
    quantize,
    sha256sum,
)

NODE_BASE_ATTRIBUTES = (
    "_children",
    "_depth",
    "_parent",
    "ast_type",
    "node_id",
    "_metadata",
    "_original_node",
    "_cache_descendants",
)
NODE_SRC_ATTRIBUTES = (
    "col_offset",
    "end_col_offset",
    "end_lineno",
    "full_source_code",
    "lineno",
    "node_source_code",
    "src",
)

DICT_AST_SKIPLIST = ("full_source_code", "node_source_code")


def get_node(
    ast_struct: Union[dict, python_ast.AST], parent: Optional["VyperNode"] = None
) -> "VyperNode":
    """
    Convert an AST structure to a vyper AST node. Entry point to constructing
    vyper AST nodes.

    This is a recursive call, all child nodes of the input value are also
    converted to Vyper nodes.

    Parameters
    ----------
    ast_struct: dict | AST
        Annotated python AST node or vyper AST dict to generate the node from.
    parent: VyperNode, optional
        Parent node of the node being created.

    Returns
    -------
    VyperNode
        The generated AST object.
    """
    if not isinstance(ast_struct, dict):
        ast_struct = ast_struct.__dict__

        # workaround: some third party module (ex. ipython) might insert
        # a "parent" member into the node, creating a duplicate kwarg
        # error below when calling vy_class()
        if "parent" in ast_struct:
            ast_struct = copy.copy(ast_struct)
            del ast_struct["parent"]

    if ast_struct["ast_type"] == "AnnAssign" and isinstance(parent, Module):
        # Replace `implements` interface declarations `AnnAssign` with `ImplementsDecl`
        if getattr(ast_struct["target"], "id", None) == "implements":
            ast_struct["ast_type"] = "ImplementsDecl"

        # Replace "uses:" `AnnAssign` nodes with `UsesDecl`
        elif getattr(ast_struct["target"], "id", None) == "uses":
            ast_struct["ast_type"] = "UsesDecl"

        # Replace "initializes:" `AnnAssign` nodes with `InitializesDecl`
        elif getattr(ast_struct["target"], "id", None) == "initializes":
            ast_struct["ast_type"] = "InitializesDecl"

        # Replace "exports:" `AnnAssign` nodes with `ExportsDecl`
        elif getattr(ast_struct["target"], "id", None) == "exports":
            ast_struct["ast_type"] = "ExportsDecl"

        # Replace state and local variable declarations `AnnAssign` with `VariableDecl`
        else:
            ast_struct["ast_type"] = "VariableDecl"

    enum_warn = False
    if ast_struct["ast_type"] == "EnumDef":
        enum_warn = True
        ast_struct["ast_type"] = "FlagDef"

    vy_class = getattr(sys.modules[__name__], ast_struct["ast_type"], None)

    if vy_class is None:
        if ast_struct["ast_type"] == "Delete":
            _raise_syntax_exc("Deleting is not supported", ast_struct)
        elif ast_struct["ast_type"] in ("ExtSlice", "Slice"):
            _raise_syntax_exc("Vyper does not support slicing", ast_struct)
        elif ast_struct["ast_type"] == "UAdd":
            _raise_syntax_exc("Vyper does not support + as a unary operator", parent)
        else:
            _raise_syntax_exc(
                f"Invalid syntax (unsupported '{ast_struct['ast_type']}' Python AST node)",
                ast_struct,
            )

    node = vy_class(parent=parent, **ast_struct)

    # TODO: Putting this after node creation to pretty print, remove after enum deprecation
    if enum_warn:
        # TODO: hack to pretty print, logic should be factored out of exception
        pretty_printed_node = str(VyperException("", node))
        warnings.warn(
            f"enum will be deprecated in a future release, use flag instead. {pretty_printed_node}",
            stacklevel=2,
        )

    node.validate()

    return node


def _to_node(obj, parent):
    # if object is a Python node or dict representing a node, convert to a Vyper node
    if isinstance(obj, (dict, python_ast.AST)):
        return get_node(obj, parent)
    if isinstance(obj, VyperNode):
        # if object is already a vyper node, make sure the parent is set correctly
        # and fix any missing source offsets
        obj.set_parent(parent)
        for field_name in NODE_SRC_ATTRIBUTES:
            if getattr(obj, field_name) is None:
                setattr(obj, field_name, getattr(parent, field_name, None))
    return obj


def _to_dict(value):
    # if value is a Vyper node, convert to a dict
    if isinstance(value, VyperNode):
        return value.to_dict()
    return value


def _node_filter(node, filters):
    # recursive equality check for VyperNode.get_children filters
    if not filters:
        return True
    for key, value in filters.items():
        if isinstance(value, set):
            if node.get(key) not in value:
                return False
        elif node.get(key) != value:
            return False
    return True


def _apply_filters(node_iter, node_type, filters, reverse):
    ret = node_iter
    if node_type is not None:
        ret = (i for i in ret if isinstance(i, node_type))
    if filters is not None:
        ret = (i for i in ret if _node_filter(i, filters))

    ret = list(ret)
    if reverse:
        ret.reverse()
    return ret


def _raise_syntax_exc(error_msg: str, ast_struct: dict) -> None:
    # helper function to raise a SyntaxException from a dict representing a node
    raise SyntaxException(
        error_msg,
        ast_struct.get("full_source_code"),
        ast_struct.get("lineno"),
        ast_struct.get("col_offset"),
    )


class VyperNode:
    """
    Base class for all vyper AST nodes.

    Vyper nodes are generated from, and closely resemble, their Python counterparts.
    Divergences are always handled in a node's `__init__` method, and explained
    in the node docstring.

    Class Attributes
    ----------------
    __slots__ : Tuple
        Allowed field names for the node.
    _description : str, optional
        A human-readable description of the node. Used to give more verbose error
        messages.
    _only_empty_fields : Tuple, optional
        Field names that, if present, must be set to None or a `SyntaxException`
        is raised. This attribute is used to exclude syntax that is valid in Python
        but not in Vyper.
    _translated_fields : Dict, optional
        Field names that are reassigned if encountered. Used to normalize fields
        across different Python versions.
    """

    __slots__ = NODE_BASE_ATTRIBUTES + NODE_SRC_ATTRIBUTES

    _public_slots = [i for i in __slots__ if not i.startswith("_")]
    _only_empty_fields: tuple = ()
    _translated_fields: dict = {}

    def __init__(self, parent: Optional["VyperNode"] = None, **kwargs: dict):
        # this function is performance-sensitive
        """
        AST node initializer method.

        Node objects are not typically instantiated directly, you should instead
        create them using the `get_node` method.

        Parameters
        ----------
        parent: VyperNode, optional
            Node which contains this node.
        **kwargs : dict
            Dictionary of fields to be included within the node.
        """
        self.set_parent(parent)
        self._children: list = []
        self._metadata: NodeMetadata = NodeMetadata()
        self._original_node = None
        self._cache_descendants = None

        for field_name in NODE_SRC_ATTRIBUTES:
            # when a source offset is not available, use the parent's source offset
            value = kwargs.pop(field_name, None)
            if value is None:
                value = getattr(parent, field_name, None)
            setattr(self, field_name, value)

        for field_name, value in kwargs.items():
            if field_name in self._translated_fields:
                field_name = self._translated_fields[field_name]

            if field_name in self.get_fields():
                if isinstance(value, list):
                    value = [_to_node(i, self) for i in value]
                else:
                    value = _to_node(value, self)
                setattr(self, field_name, value)

            elif value and field_name in self._only_empty_fields:
                _raise_syntax_exc(
                    f"Syntax is valid Python but not valid for Vyper\n"
                    f"class: {type(self).__name__}, field_name: {field_name}",
                    kwargs,
                )

        # add to children of parent last to ensure an accurate hash is generated
        if parent is not None:
            parent._children.append(self)

    @property
    def parent(self):
        return self._parent

    # set parent, can be useful when inserting copied nodes into the AST
    def set_parent(self, parent: "VyperNode"):
        self._parent = parent
        self._depth = getattr(parent, "_depth", -1) + 1

    @classmethod
    def from_node(cls, node: "VyperNode", **kwargs) -> "VyperNode":
        """
        Return a new VyperNode based on an existing node.

        This method creates a new node with the same source offsets as an existing
        node. The new node can then replace the existing node within the AST.
        Preserving source offsets ensures accurate error reporting and source
        map generation from the modified AST.

        Arguments
        ---------
        node: VyperNode
            An existing Vyper node. The generated node will have the same source
            offsets and ID as this node.
        **kwargs : Any
            Fields and values for the new node.

        Returns
        -------
        Vyper node instance
        """
        ast_struct = {i: getattr(node, i) for i in VyperNode._public_slots}
        ast_struct.update(ast_type=cls.__name__, **kwargs)
        return cls(**ast_struct)

    @classmethod
    @functools.lru_cache(maxsize=None)
    def get_fields(cls) -> set:
        """
        Return a set of field names for this node.

        Attributes that are prepended with an underscore are considered private
        and are not included within this sequence.
        """
        slot_fields = [x for i in cls.__mro__ for x in getattr(i, "__slots__", [])]
        return set(i for i in slot_fields if not i.startswith("_"))

    def __hash__(self):
        values = [getattr(self, i, None) for i in VyperNode._public_slots]
        return hash(tuple(values))

    def __deepcopy__(self, memo):
        # default implementation of deepcopy is a hotspot
        return pickle.loads(pickle.dumps(self))

    def __eq__(self, other):
        # CMC 2024-03-03 I'm not sure it makes much sense to compare AST
        # nodes, especially if they come from other modules
        if not isinstance(other, type(self)):
            return False
        if getattr(other, "node_id", None) != getattr(self, "node_id", None):
            return False
        for field_name in (i for i in self.get_fields() if i not in VyperNode.__slots__):
            if getattr(self, field_name, None) != getattr(other, field_name, None):
                return False
        return True

    def __repr__(self):
        cls = type(self)
        class_repr = f"{cls.__module__}.{cls.__qualname__}"
        return f"{class_repr}:\n{self._annotated_source}"

    @property
    def _annotated_source(self):
        # return source with context / line/col info
        return annotate_source_code(
            self.full_source_code,
            self.lineno,
            self.col_offset,
            context_lines=VYPER_ERROR_CONTEXT_LINES,
            line_numbers=VYPER_ERROR_LINE_NUMBERS,
        )

    @property
    def description(self):
        """
        Property method providing a human-readable description of a node.

        Node-specific description strings are added via the `_descrption` class
        attribute. If this attribute is not found, the name of the class is
        returned instead.
        """
        return getattr(self, "_description", type(self).__name__)

    @property
    def module_node(self):
        if isinstance(self, Module):
            return self
        return self.get_ancestor(Module)

    def get_id_dict(self):
        source_id = None
        if self.module_node is not None:
            source_id = self.module_node.source_id
        return {"node_id": self.node_id, "source_id": source_id}

    @property
    def is_literal_value(self):
        """
        Check if the node is a literal value.
        """
        return False

    @property
    def is_terminus(self):
        """
        Check if execution halts upon reaching this node.
        """
        return False

    @property
    def has_folded_value(self):
        """
        Property method to check if the node has a folded value.
        """
        return "folded_value" in self._metadata

    def get_folded_value(self) -> "ExprNode":
        """
        Attempt to get the folded value, bubbling up UnfoldableNode if the node
        is not foldable.
        """
        try:
            return self._metadata["folded_value"]
        except KeyError:
            raise UnfoldableNode("not foldable", self)

    def reduced(self) -> "ExprNode":
        if self.has_folded_value:
            return self.get_folded_value()
        return self

    def _set_folded_value(self, node: "VyperNode") -> None:
        # sanity check this is only called once
        assert "folded_value" not in self._metadata

        # set the "original node" so that exceptions can point to the original
        # node and not the folded node
        cls = node.__class__
        # make a fresh copy so that the node metadata is fresh.
        node = cls(**{i: getattr(node, i) for i in node.get_fields() if hasattr(node, i)})
        node._original_node = self

        self._metadata["folded_value"] = node

    def get_original_node(self) -> "VyperNode":
        return self._original_node or self

    def validate(self) -> None:
        """
        Validate the content of a node.

        Called by `ast.validation.validate_literal_nodes` to verify values
        within literal nodes.

        Returns `None` if the node is valid, raises `InvalidLiteral` or another
        more expressive exception if the value cannot be valid within a Vyper
        contract.
        """
        pass

    def to_dict(self) -> dict:
        """
        Return the node as a dict. Child nodes and their descendants are also converted.
        """
        ast_dict = {}
        for key in [i for i in self.get_fields() if i not in DICT_AST_SKIPLIST]:
            value = getattr(self, key, None)
            if isinstance(value, list):
                ast_dict[key] = [_to_dict(i) for i in value]
            else:
                ast_dict[key] = _to_dict(value)

        # TODO: add full analysis result, e.g. expr_info
        if "type" in self._metadata:
            ast_dict["type"] = self._metadata["type"].to_dict()

        return ast_dict

    def get_ancestor(self, node_type: Union["VyperNode", tuple, None] = None) -> "VyperNode":
        """
        Return an ancestor node for this node.

        An ancestor is any node which exists within the AST above the given node.

        Arguments
        ---------
        node_type : VyperNode | tuple, optional
            A node type or tuple of types. If given, this method checks all
            ancestor nodes of this node starting with the parent, and returns
            the first node with a type matching the given value.

        Returns
        -------
        With no arguments given: the parent of this node.

        With `node_type`: the first matching ascendant node, or `None` if no node
        is found which matches the argument value.
        """
        if node_type is None or self._parent is None:
            return self._parent

        if isinstance(self._parent, node_type):
            return self._parent

        return self._parent.get_ancestor(node_type)

    def get_children(
        self,
        node_type: Union["VyperNode", tuple, None] = None,
        filters: Optional[dict] = None,
        reverse: bool = False,
    ) -> list:
        """
        Return a list of children of this node which match the given filter(s).

        Results are sorted by the starting source offset and node ID, ascending.

        Parameters
        ----------
        node_type : VyperNode | tuple, optional
            A node type or tuple of types. If given, only child nodes where the
            type matches this value are returned. This is functionally identical
            to calling `isinstance(child, node_type)`
        filters : dict, optional
            Dictionary of attribute names and expected values. Only nodes that
            contain the given attributes and match the given values are returned.
            * You can use dots within the name in order to check members of members.
              e.g. `{'annotation.func.id': "constant"}`
            * Expected values may be given as a set, in order to match a node must
              contain the given attribute and match any one value within the set.
              e.g. `{'id': {'public', 'constant'}}` will match nodes with an `id`
                    member that contains either "public" or "constant".
        reverse : bool, optional
            If `True`, the order of results is reversed prior to return.

        Returns
        -------
        list
            Child nodes matching the filter conditions.
        """
        return _apply_filters(iter(self._children), node_type, filters, reverse)

    def get_descendants(
        self,
        node_type: Union["VyperNode", tuple, None] = None,
        filters: Optional[dict] = None,
        include_self: bool = False,
        reverse: bool = False,
    ) -> list:
        # this function is performance-sensitive
        """
        Return a list of descendant nodes of this node which match the given filter(s).

        A descendant is any node which exists within the AST beneath the given node.

        Results are sorted by the starting source offset and depth, ascending. You
        can rely on that the sequence will always contain a parent node prior to any
        of it's children. If the result is reversed, all children of a node will
        be in the sequence prior to their parent.

        Parameters
        ----------
        node_type : VyperNode | tuple, optional
            A node type or tuple of types. If given, only child nodes where the
            type matches this value are returned. This is functionally identical
            to calling `isinstance(child, node_type)`
        filters : dict, optional
            Dictionary of attribute names and expected values. Only nodes that
            contain the given attributes and match the given values are returned.
            * You can use dots within the name in order to check members of members.
              e.g. `{'annotation.func.id': "constant"}`
            * Expected values may be given as a set, in order to match a node must
              contain the given attribute and match any one value within the set.
              e.g. `{'id': {'public', 'constant'}}` will match nodes with an `id`
                    member that contains either "public" or "constant".
        include_self : bool, optional
            If True, this node is also included in the search results if it matches
            the given filter.
        reverse : bool, optional
            If `True`, the order of results is reversed prior to return.

        Returns
        -------
        list
            Descendant nodes matching the filter conditions.
        """
        ret = self._get_descendants(include_self)
        return _apply_filters(ret, node_type, filters, reverse)

    def _get_descendants(self, include_self=True):
        # get descendants in topsort order
        if self._cache_descendants is None:
            ret = [self]
            for node in self._children:
                ret.extend(node._get_descendants())

            self._cache_descendants = ret

        ret = iter(self._cache_descendants)

        if not include_self:
            s = next(ret)  # pop
            assert s is self

        return ret

    def get(self, field_str: str) -> Any:
        """

        Recursive getter function for node attributes.

        Parameters
        ----------
        field_str : str
            Attribute string of the location of the node to return.

        Returns
        -------
        VyperNode : optional
            Value at the location of the given field string, if one
            exists. `None` if the field string is empty or invalid.
        """
        obj = self
        for key in field_str.split("."):
            obj = getattr(obj, key, None)
        return obj


class TopLevel(VyperNode):
    """
    Inherited class for Module and FunctionDef nodes.

    Class attributes
    ----------------
    doc_string : Expr
        Expression node representing the docstring within this node.
    """

    __slots__ = ("body", "name", "doc_string")


class Module(TopLevel):
    # metadata
    __slots__ = ("path", "resolved_path", "source_id")

    def to_dict(self):
        return dict(source_sha256sum=self.source_sha256sum, **super().to_dict())

    @property
    def source_sha256sum(self):
        return sha256sum(self.full_source_code)

    @contextlib.contextmanager
    def namespace(self):
        from vyper.semantics.namespace import get_namespace, override_global_namespace

        # kludge implementation for backwards compatibility.
        # TODO: replace with type_from_ast
        try:
            ns = self._metadata["namespace"]
        except AttributeError:
            ns = get_namespace()
        with override_global_namespace(ns):
            yield


class FunctionDef(TopLevel):
    __slots__ = ("args", "returns", "decorator_list", "pos")


class DocStr(VyperNode):
    """
    A docstring.

    Attributes
    ----------
    value : str
        Value of the node, represented as an string.
    """

    __slots__ = ("value",)
    _translated_fields = {"s": "value"}


class arguments(VyperNode):
    __slots__ = ("args", "defaults", "default")
    _only_empty_fields = ("posonlyargs", "vararg", "kwonlyargs", "kwarg", "kw_defaults")


class arg(VyperNode):
    __slots__ = ("arg", "annotation")


# base class for stmt nodes. doesn't do anything except classification
class Stmt(VyperNode):
    pass


class Return(Stmt):
    __slots__ = ("value",)

    @property
    def is_terminus(self):
        return True


class Expr(Stmt):
    __slots__ = ("value",)

    @property
    def is_terminus(self):
        return self.value.is_terminus


class NamedExpr(Stmt):
    __slots__ = ("target", "value")

    def validate(self):
        # module[dep1 := dep2]

        # XXX: better error messages
        if not isinstance(self.target, Name):
            raise StructureException("not a Name")

        if not isinstance(self.value, Name):
            raise StructureException("not a Name")


class Log(Stmt):
    __slots__ = ("value",)

    def validate(self):
        if not isinstance(self.value, Call):
            raise StructureException("Log must call an event", self.value)


class FlagDef(TopLevel):
    __slots__ = ("name", "body")


class EventDef(TopLevel):
    __slots__ = ("name", "body")


class InterfaceDef(TopLevel):
    __slots__ = ("name", "body")


class StructDef(TopLevel):
    __slots__ = ("name", "body")


# base class for expression nodes
# note that it is named ExprNode to avoid a conflict with
# the Expr type (which is a type of statement node, see python AST docs).
class ExprNode(VyperNode):
    __slots__ = ("_expr_info",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._expr_info = None

    def to_dict(self):
        ret = super().to_dict()
        if self._expr_info is None:
            return ret

        reads = [s.to_dict() for s in self._expr_info._reads]
        reads = [s for s in reads if s]
        if reads:
            ret["variable_reads"] = reads

        writes = [s.to_dict() for s in self._expr_info._writes]
        writes = [s for s in writes if s]
        if writes:
            ret["variable_writes"] = writes

        return ret


class Constant(ExprNode):
    # inherited class for all simple constant node types
    __slots__ = ("value",)

    @property
    def is_literal_value(self):
        return True


class Num(Constant):
    # inherited class for all numeric constant node types
    __slots__ = ()

    def validate(self):
        if self.value < SizeLimits.MIN_INT256:
            raise OverflowException("Value is below lower bound for all numeric types", self)
        if self.value > SizeLimits.MAX_UINT256:
            raise OverflowException("Value exceeds upper bound for all numeric types", self)


class Int(Num):
    """
    An integer.

    Attributes
    ----------
    value : int
        Value of the node, represented as an integer.
    """

    __slots__ = ()


class Decimal(Num):
    """
    A decimal.

    Attributes
    ----------
    value : decimal.Decimal
        Value of the node, represented as a Decimal object.
    """

    __slots__ = ()

    def __init__(self, parent: Optional["VyperNode"] = None, **kwargs: dict):
        super().__init__(parent, **kwargs)
        if not isinstance(self.value, decimal.Decimal):
            self.value = decimal.Decimal(self.value)

    def to_dict(self):
        ast_dict = super().to_dict()
        ast_dict["value"] = self.node_source_code
        return ast_dict

    def validate(self):
        # note: maybe use self.value == quantize(self.value) for this check
        if self.value.as_tuple().exponent < -MAX_DECIMAL_PLACES:
            raise InvalidLiteral("Vyper supports a maximum of ten decimal points", self)
        if self.value < SizeLimits.MIN_AST_DECIMAL:
            raise OverflowException("Value is below lower bound for decimal types", self)
        if self.value > SizeLimits.MAX_AST_DECIMAL:
            raise OverflowException("Value exceeds upper bound for decimal types", self)


class Hex(Constant):
    """
    A hexadecimal value, e.g. `0xFF`

    Attributes
    ----------
    value : str
        Value of the node, represented as a string taken directly from the contract source.
    """

    __slots__ = ()

    def validate(self):
        if "_" in self.value:
            raise InvalidLiteral("Underscores not allowed in hex literals", self)
        if len(self.value) % 2:
            raise InvalidLiteral("Hex notation requires an even number of digits", self)

    @property
    def n_nibbles(self):
        """
        The number of nibbles this hex value represents
        """
        return len(self.value) - 2

    @property
    def n_bytes(self):
        """
        The number of bytes this hex value represents
        """
        return len(self.bytes_value)

    @property
    def bytes_value(self):
        """
        This value as bytes
        """
        return bytes.fromhex(self.value.removeprefix("0x"))


class Str(Constant):
    __slots__ = ()
    _translated_fields = {"s": "value"}

    def validate(self):
        for c in self.value:
            if ord(c) >= 256:
                raise InvalidLiteral(f"'{c}' is not an allowed string literal character", self)


class Bytes(Constant):
    __slots__ = ()
    _translated_fields = {"s": "value"}

    def __init__(self, parent: Optional["VyperNode"] = None, **kwargs: dict):
        super().__init__(parent, **kwargs)
        if isinstance(self.value, str):
            # convert hex string to bytes
            length = len(self.value) // 2 - 1
            self.value = int(self.value, 16).to_bytes(length, "big")

    def to_dict(self):
        ast_dict = super().to_dict()
        ast_dict["value"] = f"0x{self.value.hex()}"
        return ast_dict

    @property
    def s(self):
        return self.value


class List(ExprNode):
    __slots__ = ("elements",)
    _translated_fields = {"elts": "elements"}

    @property
    def is_literal_value(self):
        return all(e.is_literal_value for e in self.elements)


class Tuple(ExprNode):
    __slots__ = ("elements",)
    _translated_fields = {"elts": "elements"}

    @property
    def is_literal_value(self):
        return all(e.is_literal_value for e in self.elements)

    def validate(self):
        if not self.elements:
            raise InvalidLiteral("Cannot have an empty tuple", self)


class NameConstant(Constant):
    __slots__ = ()

    def validate(self):
        if self.value is None:
            raise InvalidLiteral("`None` is not a valid vyper value!", self)


class Ellipsis(Constant):
    __slots__ = ()


class Dict(ExprNode):
    __slots__ = ("keys", "values")

    @property
    def is_literal_value(self):
        return all(v.is_literal_value for v in self.values)


class Name(ExprNode):
    __slots__ = ("id",)


class UnaryOp(ExprNode):
    __slots__ = ("op", "operand")


class Operator(VyperNode):
    pass


class USub(Operator):
    __slots__ = ()
    _description = "negation"
    _op = operator.neg


class Not(Operator):
    __slots__ = ()
    _op = operator.not_


class Invert(Operator):
    __slots__ = ()
    _description = "bitwise not"
    _pretty = "~"

    def _op(self, value):
        return (2**256 - 1) ^ value


class BinOp(ExprNode):
    __slots__ = ("left", "op", "right")


class Add(Operator):
    __slots__ = ()
    _description = "addition"
    _pretty = "+"
    _op = operator.add


class Sub(Operator):
    __slots__ = ()
    _description = "subtraction"
    _pretty = "-"
    _op = operator.sub


class Mult(Operator):
    __slots__ = ()
    _description = "multiplication"
    _pretty = "*"

    def _op(self, left, right):
        assert type(left) is type(right)
        value = left * right
        if isinstance(left, decimal.Decimal):
            # ensure that the result is truncated to MAX_DECIMAL_PLACES
            try:
                # if the intermediate result requires too many decimal places,
                # decimal will puke - catch the error and raise an
                # OverflowException
                return quantize(value)
            except decimal.InvalidOperation:
                msg = f"{self._description} requires too many decimal places:"
                msg += f"\n  {left} * {right} => {value}"
                raise OverflowException(msg, self) from None
        else:
            return value


class Div(Operator):
    __slots__ = ()
    _description = "decimal division"
    _pretty = "/"

    def _op(self, left, right):
        # evaluate the operation using true division or floor division
        assert type(left) is type(right)
        if not right:
            raise ZeroDivisionException("Division by zero")

        if not isinstance(left, decimal.Decimal):
            raise UnfoldableNode("Cannot use `/` on non-decimals (did you mean `//`?)")

        value = left / right
        if value < 0:
            # the EVM always truncates toward zero
            value = -(-left / right)
        # ensure that the result is truncated to MAX_DECIMAL_PLACES
        try:
            return quantize(value)
        except decimal.InvalidOperation:
            msg = f"{self._description} requires too many decimal places:"
            msg += f"\n  {left} {self._pretty} {right} => {value}"
            raise OverflowException(msg, self) from None


class FloorDiv(VyperNode):
    __slots__ = ()
    _description = "integer division"
    _pretty = "//"

    def _op(self, left, right):
        # evaluate the operation using true division or floor division
        assert type(left) is type(right)
        if not right:
            raise ZeroDivisionException("Division by zero")

        if not isinstance(left, int):
            raise UnfoldableNode("Cannot use `//` on non-integers (did you mean `/`?)")

        return evm_div(left, right)


class Mod(Operator):
    __slots__ = ()
    _description = "modulus"
    _pretty = "%"

    def _op(self, left, right):
        if not right:
            raise ZeroDivisionException("Modulo by zero")

        value = abs(left) % abs(right)
        if left < 0:
            value = -value
        return value


class Pow(Operator):
    __slots__ = ()
    _description = "exponentiation"
    _pretty = "**"

    def _op(self, left, right):
        if isinstance(left, decimal.Decimal):
            raise TypeMismatch("Cannot perform exponentiation on decimal values.", self._parent)
        if right < 0:
            raise InvalidOperation("Cannot calculate a negative power", self._parent)
        # prevent a compiler hang. we are ok with false positives at this
        # stage since we are just trying to filter out inputs which can cause
        # the compiler to hang. the others will get caught during constant
        # folding or codegen.
        # l**r > 2**256
        # r * ln(l) > ln(2 ** 256)
        # r > ln(2 ** 256) / ln(l)
        if right > math.log(decimal.Decimal(2**257)) / math.log(decimal.Decimal(left)):
            raise InvalidLiteral("Out of bounds", self)

        return int(left**right)


class BitAnd(Operator):
    __slots__ = ()
    _description = "bitwise and"
    _pretty = "&"
    _op = operator.and_


class BitOr(Operator):
    __slots__ = ()
    _description = "bitwise or"
    _pretty = "|"
    _op = operator.or_


class BitXor(Operator):
    __slots__ = ()
    _description = "bitwise xor"
    _pretty = "^"
    _op = operator.xor


class LShift(Operator):
    __slots__ = ()
    _description = "bitwise left shift"
    _pretty = "<<"
    _op = operator.lshift


class RShift(Operator):
    __slots__ = ()
    _description = "bitwise right shift"
    _pretty = ">>"
    _op = operator.rshift


class BoolOp(ExprNode):
    __slots__ = ("op", "values")


class And(Operator):
    __slots__ = ()
    _description = "logical and"
    _op = all


class Or(Operator):
    __slots__ = ()
    _description = "logical or"
    _op = any


class Compare(ExprNode):
    """
    A comparison of two values.

    Attributes
    ----------
    left : ExprNode
        The left-hand value in the comparison.
    op : Operator
        The comparison operator.
    right : ExprNode
        The right-hand value in the comparison.
    """

    __slots__ = ("left", "op", "right")

    def __init__(self, *args, **kwargs):
        if len(kwargs["ops"]) > 1 or len(kwargs["comparators"]) > 1:
            _raise_syntax_exc("Cannot have a comparison with more than two elements", kwargs)

        kwargs["op"] = kwargs.pop("ops")[0]
        kwargs["right"] = kwargs.pop("comparators")[0]
        super().__init__(*args, **kwargs)


class Eq(Operator):
    __slots__ = ()
    _description = "equality"
    _op = operator.eq


class NotEq(Operator):
    __slots__ = ()
    _description = "non-equality"
    _op = operator.ne


class Lt(Operator):
    __slots__ = ()
    _description = "less than"
    _op = operator.lt


class LtE(Operator):
    __slots__ = ()
    _description = "less-or-equal"
    _op = operator.le


class Gt(Operator):
    __slots__ = ()
    _description = "greater than"
    _op = operator.gt


class GtE(Operator):
    __slots__ = ()
    _description = "greater-or-equal"
    _op = operator.ge


class In(Operator):
    __slots__ = ()
    _description = "membership"

    def _op(self, left, right):
        return left in right


class NotIn(Operator):
    __slots__ = ()
    _description = "exclusion"

    def _op(self, left, right):
        return left not in right


class Call(ExprNode):
    __slots__ = ("func", "args", "keywords")

    @property
    def is_extcall(self):
        return isinstance(self._parent, ExtCall)

    @property
    def is_staticcall(self):
        return isinstance(self._parent, StaticCall)

    @property
    def is_plain_call(self):
        return not (self.is_extcall or self.is_staticcall)

    @property
    def kind_str(self):
        if self.is_extcall:
            return "extcall"
        if self.is_staticcall:
            return "staticcall"
        raise CompilerPanic("unreachable!")  # pragma: nocover

    @property
    def is_terminus(self):
        # cursed import cycle!
        from vyper.builtins.functions import get_builtin_functions

        if not isinstance(self.func, Name):
            return False

        funcname = self.func.id
        builtin_t = get_builtin_functions().get(funcname)
        if builtin_t is None:
            return False

        return builtin_t._is_terminus


class ExtCall(ExprNode):
    __slots__ = ("value",)

    def validate(self):
        if not isinstance(self.value, Call):
            # TODO: investigate wrong col_offset for `self.value`
            raise StructureException(
                "`extcall` must be followed by a function call",
                self.value,
                hint="did you forget parentheses?",
            )


class StaticCall(ExprNode):
    __slots__ = ("value",)

    def validate(self):
        if not isinstance(self.value, Call):
            raise StructureException(
                "`staticcall` must be followed by a function call",
                self.value,
                hint="did you forget parentheses?",
            )


class keyword(VyperNode):
    __slots__ = ("arg", "value")


class Attribute(ExprNode):
    __slots__ = ("attr", "value")


class Subscript(ExprNode):
    __slots__ = ("slice", "value")


class Assign(Stmt):
    """
    An assignment.

    Attributes
    ----------
    target : VyperNode
        Left-hand side of the assignment.
    value : ExprNode
        Right-hand side of the assignment.
    """

    __slots__ = ("target", "value")

    def __init__(self, *args, **kwargs):
        if len(kwargs["targets"]) > 1:
            _raise_syntax_exc("Assignment statement must have one target", kwargs)

        kwargs["target"] = kwargs.pop("targets")[0]
        super().__init__(*args, **kwargs)


class AnnAssign(VyperNode):
    __slots__ = ("target", "annotation", "value")


class VariableDecl(VyperNode):
    """
    A contract variable declaration.

    Excludes `simple` attribute from Python `AnnAssign` node.

    Attributes
    ----------
    target : VyperNode
        Left-hand side of the assignment.
    value : VyperNode
        Right-hand side of the assignment.
    annotation : VyperNode
        Type of variable.
    is_constant : bool, optional
        If true, indicates that the variable is a constant variable.
    is_public : bool, optional
        If true, indicates that the variable is a public state variable.
    is_immutable : bool, optional
        If true, indicates that the variable is an immutable variable.
    """

    __slots__ = (
        "target",
        "annotation",
        "value",
        "is_constant",
        "is_public",
        "is_immutable",
        "is_transient",
        "_expanded_getter",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_constant = False
        self.is_public = False
        self.is_immutable = False
        self.is_transient = False
        self._expanded_getter = None

        def _check_args(annotation, call_name):
            # do the same thing as `validate_call_args`
            # (can't be imported due to cyclic dependency)
            if len(annotation.args) != 1:
                raise ArgumentException(f"Invalid number of arguments to `{call_name}`:", self)

        # the annotation is a "function" call, e.g.
        # `foo: public(constant(uint256))`
        # pretend we were parsing actual Vyper AST. annotation would be
        # TYPE | PUBLIC "(" TYPE | ((IMMUTABLE | CONSTANT) "(" TYPE ")") ")"
        if self.annotation.get("func.id") == "public":
            _check_args(self.annotation, "public")
            self.is_public = True
            # unwrap one layer
            self.annotation = self.annotation.args[0]

        func_id = self.annotation.get("func.id")
        if func_id in ("immutable", "constant", "transient"):
            _check_args(self.annotation, func_id)
            setattr(self, f"is_{func_id}", True)
            # unwrap one layer
            self.annotation = self.annotation.args[0]

        if isinstance(self.annotation, Call):
            _raise_syntax_exc("Invalid scope for variable declaration", self.annotation)

    @property
    def _pretty_location(self) -> str:
        if self.is_constant:
            return "Constant"
        if self.is_transient:
            return "Transient"
        if self.is_immutable:
            return "Immutable"
        return "Storage"

    def validate(self):
        if self.is_constant and self.value is None:
            raise VariableDeclarationException("Constant must be declared with a value", self)

        if not self.is_constant and self.value is not None:
            raise VariableDeclarationException(
                f"{self._pretty_location} variables cannot have an initial value", self.value
            )
        if not isinstance(self.target, Name):
            raise VariableDeclarationException("Invalid variable declaration", self.target)


class AugAssign(Stmt):
    __slots__ = ("op", "target", "value")


class Raise(Stmt):
    __slots__ = ("exc",)
    _only_empty_fields = ("cause",)

    @property
    def is_terminus(self):
        return True


class Assert(Stmt):
    __slots__ = ("test", "msg")


class Pass(Stmt):
    __slots__ = ()


class _ImportStmt(Stmt):
    __slots__ = ("name", "alias")

    def to_dict(self):
        ret = super().to_dict()
        if (import_info := self._metadata.get("import_info")) is not None:
            ret["import_info"] = import_info.to_dict()

        return ret

    def __init__(self, *args, **kwargs):
        if len(kwargs["names"]) > 1:
            _raise_syntax_exc("Assignment statement must have one target", kwargs)
        names = kwargs.pop("names")[0]
        kwargs["name"] = names.name
        kwargs["alias"] = names.asname
        super().__init__(*args, **kwargs)


class Import(_ImportStmt):
    __slots__ = ()


class ImportFrom(_ImportStmt):
    __slots__ = ("level", "module")


class ImplementsDecl(Stmt):
    """
    An `implements` declaration.

    Attributes
    ----------
    annotation : Name
        Name node for the interface to be implemented
    """

    __slots__ = ("annotation",)
    _only_empty_fields = ("value",)

    def validate(self):
        if not isinstance(self.annotation, (Name, Attribute)):
            raise StructureException("invalid implements", self.annotation)


def as_tuple(node: VyperNode):
    """
    Convenience function for some AST nodes which allow either a Tuple
    or single elements. Returns a python tuple of AST nodes.
    """
    if isinstance(node, Tuple):
        return node.elements
    else:
        return (node,)


class UsesDecl(Stmt):
    """
    A `uses` declaration.

    Attributes
    ----------
    annotation : Name | Attribute | Tuple
        The module(s) which this uses
    """

    __slots__ = ("annotation",)
    _only_empty_fields = ("value",)

    def validate(self):
        items = as_tuple(self.annotation)
        for item in items:
            if not isinstance(item, (Name, Attribute)):
                raise StructureException("invalid uses", item)


class InitializesDecl(Stmt):
    """
    An `initializes` declaration.

    Attributes
    ----------
    annotation : Name | Attribute | Subscript
        An imported module which this module initializes
    """

    __slots__ = ("annotation",)
    _only_empty_fields = ("value",)

    def validate(self):
        module_ref = self.annotation
        if isinstance(module_ref, Subscript):
            dependencies = as_tuple(module_ref.slice)
            module_ref = module_ref.value

            for item in dependencies:
                if not isinstance(item, NamedExpr):
                    raise StructureException(
                        "invalid dependency (hint: should be [dependency := dependency]", item
                    )
                if not isinstance(item.target, (Name, Attribute)):
                    raise StructureException("invalid module", item.target)
                if not isinstance(item.value, (Name, Attribute)):
                    raise StructureException("invalid module", item.target)

        if not isinstance(module_ref, (Name, Attribute)):
            raise StructureException("invalid module", module_ref)


class ExportsDecl(Stmt):
    """
    An `exports` declaration.

    Attributes
    ----------
    annotation : Name | Attribute | Tuple
        List of exports
    """

    __slots__ = ("annotation",)
    _only_empty_fields = ("value",)

    def validate(self):
        items = as_tuple(self.annotation)
        for item in items:
            if not isinstance(item, (Name, Attribute)):
                raise StructureException("invalid exports", item)


class If(Stmt):
    __slots__ = ("test", "body", "orelse")


class IfExp(ExprNode):
    __slots__ = ("test", "body", "orelse")


class For(Stmt):
    __slots__ = ("target", "iter", "body")
    _only_empty_fields = ("orelse",)


class Break(Stmt):
    __slots__ = ()


class Continue(Stmt):
    __slots__ = ()



from vyper.utils import OrderedSet
from vyper.venom.basicblock import IRBasicBlock
from vyper.venom.passes.base_pass import IRPass


class StackReorderPass(IRPass):
    visited: OrderedSet

    def _reorder_stack(self):
        pass

    def _visit(self, bb: IRBasicBlock):
        if bb in self.visited:
            return
        self.visited.add(bb)

        for bb_out in bb.cfg_out:
            self._visit(bb_out)

    def _run_pass(self):
        entry = self.function.entry
        self.visited = OrderedSet()
        self._visit(entry)



import json
from typing import Type

import pytest

from vyper import compiler
from vyper.exceptions import NamespaceCollision, StructureException, VyperException

# For reproducibility, use precompiled data of `hello: public(uint256)` using vyper 0.3.1
PRECOMPILED_ABI = """[{"stateMutability": "view", "type": "function", "name": "hello", "inputs": [], "outputs": [{"name": "", "type": "uint256"}], "gas": 2460}]"""  # noqa: E501, FS003
PRECOMPILED_BYTECODE = """0x61004456600436101561000d57610035565b60046000601c376000513461003b576319ff1d2181186100335760005460e052602060e0f35b505b60006000fd5b600080fd5b61000461004403610004600039610004610044036000f3"""  # noqa: E501
PRECOMPILED_BYTECODE_RUNTIME = """0x600436101561000d57610035565b60046000601c376000513461003b576319ff1d2181186100335760005460e052602060e0f35b505b60006000fd5b600080fd"""  # noqa: E501
PRECOMPILED = bytes.fromhex(PRECOMPILED_BYTECODE_RUNTIME[2:])


@pytest.fixture
def precompiled_contract(env):
    bytecode = bytes.fromhex(PRECOMPILED_BYTECODE.removeprefix("0x"))
    return env.deploy(json.loads(PRECOMPILED_ABI), bytecode)


@pytest.mark.parametrize(
    ("start", "length", "expected"), [(0, 5, PRECOMPILED[:5]), (5, 10, PRECOMPILED[5:][:10])]
)
def test_address_code_slice(
    start: int, length: int, expected: bytes, precompiled_contract, get_contract
):
    code = f"""
@external
def code_slice(x: address) -> Bytes[{length}]:
    return slice(x.code, {start}, {length})
"""
    contract = get_contract(code)
    actual = contract.code_slice(precompiled_contract.address)
    assert actual == expected


def test_address_code_runtime_error_slice_too_long(precompiled_contract, get_contract, tx_failed):
    start = len(PRECOMPILED) - 5
    length = 10
    code = f"""
@external
def code_slice(x: address) -> Bytes[{length}]:
    return slice(x.code, {start}, {length})
"""
    contract = get_contract(code)
    with tx_failed():
        contract.code_slice(precompiled_contract.address)


def test_address_code_runtime_error_no_code(get_contract, tx_failed):
    code = """
@external
def code_slice(x: address) -> Bytes[4]:
    return slice(x.code, 0, 4)
"""
    contract = get_contract(code)
    with tx_failed():
        contract.code_slice(b"\x00" * 20)


@pytest.mark.parametrize(
    ("bad_code", "error_type", "error_message"),
    [
        (
            # `(address).code` without `slice`
            """
@external
def code_slice(x: address) -> uint256:
    y: uint256 = convert(x.code, uint256)
    return y
""",
            StructureException,
            "(address).code is only allowed inside of a slice function with a constant length",
        ),
        (
            """
a: HashMap[Bytes[4], uint256]

@external
def foo(x: address):
    self.a[x.code] += 1
""",
            StructureException,
            "(address).code is only allowed inside of a slice function with a constant length",
        ),
        (
            # `len` not supported
            """
@external
def code_slice(x: address) -> uint256:
    y: uint256 = len(x.code)
    return y
""",
            StructureException,
            "(address).code is only allowed inside of a slice function with a constant length",
        ),
        (
            # `slice` with non static length
            """
@external
def code_slice(x: address, y: uint256) -> Bytes[4]:
    z: Bytes[4] = slice(x.code, 0, y)
    return z
""",
            StructureException,
            "(address).code is only allowed inside of a slice function with a constant length",
        ),
        (
            # `self.code` is already defined since `self` is address
            """
code: public(Bytes[4])
""",
            NamespaceCollision,
            "Member 'code' already exists in self",
        ),
    ],
)
def test_address_code_compile_error(
    bad_code: str, error_type: Type[VyperException], error_message: str
):
    with pytest.raises(error_type) as excinfo:
        compiler.compile_code(bad_code)
    assert excinfo.value.message == error_message


@pytest.mark.parametrize(
    "code",
    [
        # Environment variable
        """
@external
def foo() -> Bytes[4]:
    return slice(msg.sender.code, 0, 4)
""",
        # User defined struct
        """
struct S:
    a: address

@external
def foo(s: S) -> Bytes[4]:
    return slice(s.a.code, 0, 4)
""",
        # External contract call
        """
interface Test:
    def out_literals() -> address : view

@external
def foo(x: address) -> Bytes[4]:
    return slice((staticcall Test(x).out_literals()).code, 0, 4)
""",
    ],
)
def test_address_code_compile_success(code: str):
    compiler.compile_code(code)


def test_address_code_self_success(get_contract):
    code = """
code_deployment: public(Bytes[32])

@deploy
def __init__():
    self.code_deployment = slice(self.code, 0, 32)

@external
def code_runtime() -> Bytes[32]:
    return slice(self.code, 0, 32)
"""
    contract = get_contract(code)
    code_compiled = compiler.compile_code(code, output_formats=["bytecode", "bytecode_runtime"])
    assert contract.code_deployment() == bytes.fromhex(code_compiled["bytecode"][2:])[:32]
    assert contract.code_runtime() == bytes.fromhex(code_compiled["bytecode_runtime"][2:])[:32]


def test_address_code_self_runtime_error_deployment(get_contract, tx_failed):
    code = """
dummy: public(Bytes[1000000])

@deploy
def __init__():
    self.dummy = slice(self.code, 0, 1000000)
"""
    with tx_failed():
        get_contract(code)


def test_address_code_self_runtime_error_runtime(get_contract, tx_failed):
    code = """
@external
def code_runtime() -> Bytes[1000000]:
    return slice(self.code, 0, 1000000)
"""
    contract = get_contract(code)
    with tx_failed():
        contract.code_runtime()



import hypothesis
import pytest


@pytest.fixture(scope="module")
def little_endian_contract(get_contract):
    code = """
@internal
@view
def to_little_endian_64(_value: uint256) -> Bytes[8]:
    y: uint256 = 0
    x: uint256 = _value
    for _: uint256 in range(8):
        y = (y << 8) | (x & 255)
        x >>= 8
    return slice(convert(y, bytes32), 24, 8)

@external
@view
def get_count(counter: uint256) -> Bytes[24]:
    return self.to_little_endian_64(counter)
    """
    return get_contract(code)


@pytest.mark.fuzzing
@hypothesis.given(value=hypothesis.strategies.integers(min_value=0, max_value=(2**64 - 1)))
def test_zero_pad_range(little_endian_contract, value):
    actual_bytes = value.to_bytes(8, byteorder="little")
    contract_bytes = little_endian_contract.get_count(value)
    assert contract_bytes == actual_bytes



import math
from decimal import Decimal

from tests.utils import decimal_to_int


def test_floor(get_contract):
    code = """
x: decimal

@deploy
def __init__():
    self.x = 504.0000000001

@external
def x_floor() -> int256:
    return floor(self.x)

@external
def foo() -> int256:
    return floor(1.9999999999)

@external
def fop() -> int256:
    return floor(1.0000000001)

@external
def foq() -> int256:
    return floor(18707220957835557353007165858768422651595.9365500927)

@external
def fos() -> int256:
    return floor(0.0)

@external
def fot() -> int256:
    return floor(0.0000000001)

@external
def fou() -> int256:
    a: decimal = 305.0
    b: decimal = 100.0
    c: decimal = a / b
    return floor(c)
"""
    c = get_contract(code)
    assert c.x_floor() == 504
    assert c.foo() == 1
    assert c.fop() == 1
    assert c.foq() == math.floor(Decimal(2**167 - 1) / 10**10)
    assert c.fos() == 0
    assert c.fot() == 0
    assert c.fou() == 3


def test_floor_negative(get_contract):
    code = """
x: decimal

@deploy
def __init__():
    self.x = -504.0000000001

@external
def x_floor() -> int256:
    return floor(self.x)

@external
def foo() -> int256:
    a: int128 = -65
    b: decimal = convert(a, decimal) / 10.0
    return floor(b)

@external
def fop() -> int256:
    return floor(-27.0)

@external
def foq() -> int256:
    return floor(-9000.0000000001)

@external
def fos() -> int256:
    return floor(-0.0000000001)

@external
def fot() -> int256:
    return floor(-18707220957835557353007165858768422651595.9365500928)

@external
def fou() -> int256:
    a: decimal = -305.0
    b: decimal = 100.0
    c: decimal = a / b
    return floor(c)

@external
def floor_param(p: decimal) -> int256:
    return floor(p)
"""

    c = get_contract(code)

    assert c.x_floor() == -505
    assert c.foo() == -7
    assert c.fop() == -27
    assert c.foq() == -9001
    assert c.fos() == -1
    assert c.fot() == math.floor(-Decimal(2**167) / 10**10)
    assert c.fou() == -4
    assert c.floor_param(decimal_to_int("-5.6")) == -6
    assert c.floor_param(decimal_to_int("-0.0000000001")) == -1


def test_floor_ext_call(side_effects_contract, assert_side_effects_invoked, get_contract):
    code = """
interface Foo:
    def foo(x: decimal) -> decimal: nonpayable

@external
def foo(a: Foo) -> int256:
    return floor(extcall a.foo(2.5))
    """

    c1 = side_effects_contract("decimal")
    c2 = get_contract(code)

    assert c2.foo(c1.address) == 2

    assert_side_effects_invoked(c1, lambda: c2.foo(c1.address))


def test_floor_internal_call(get_contract):
    code = """
@external
def foo() -> int256:
    return floor(self.bar())

@internal
def bar() -> decimal:
    return 2.5
    """

    c = get_contract(code)

    assert c.foo() == 2



import itertools

import pytest

from tests.utils import decimal_to_int
from vyper.exceptions import ArrayIndexException, OverflowException, TypeMismatch


def _map_nested(f, xs):
    ret = []
    for x in xs:
        if isinstance(x, list):
            ret.append(_map_nested(f, x))
        else:
            ret.append(f(x))
    return ret


def test_list_tester_code(get_contract):
    list_tester_code = """
z: int128[3]
z2: int128[2][2]
z3: int128[2]

@external
def foo(x: int128[3]) -> int128:
    return x[0] + x[1] + x[2]

@external
def goo(x: int128[2][2]) -> int128:
    return x[0][0] + x[0][1] + x[1][0] * 10 + x[1][1] * 10

@external
def hoo(x: int128[3]) -> int128:
    y: int128[3] = x
    return y[0] + x[1] + y[2]

@external
def joo(x: int128[2][2]) -> int128:
    y: int128[2][2] = x
    y2: int128[2] = x[1]
    return y[0][0] + y[0][1] + y2[0] * 10 + y2[1] * 10

@external
def koo(x: int128[3]) -> int128:
    self.z = x
    return self.z[0] + x[1] + self.z[2]

@external
def loo(x: int128[2][2]) -> int128:
    self.z2 = x
    self.z3 = x[1]
    return self.z2[0][0] + self.z2[0][1] + self.z3[0] * 10 + self.z3[1] * 10
    """

    c = get_contract(list_tester_code)
    assert c.foo([3, 4, 5]) == 12
    assert c.goo([[1, 2], [3, 4]]) == 73
    assert c.hoo([3, 4, 5]) == 12
    assert c.joo([[1, 2], [3, 4]]) == 73
    assert c.koo([3, 4, 5]) == 12
    assert c.loo([[1, 2], [3, 4]]) == 73
    print("Passed list tests")


def test_list_output_tester_code(get_contract):
    list_output_tester_code = """
z: int128[2]

@external
def foo() -> int128[2]:
    return [3, 5]

@external
def goo() -> int128[2]:
    x: int128[2] = [3, 5]
    return x

@external
def hoo() -> int128[2]:
    self.z = [3, 5]
    return self.z

@external
def joo() -> int128[2]:
    self.z = [3, 5]
    x: int128[2] = self.z
    return x

@external
def koo() -> int128[2][2]:
    return [[1, 2], [3, 4]]

@external
def loo() -> int128[2][2]:
    x: int128[2][2] = [[1, 2], [3, 4]]
    return x

@external
def moo() -> int128[2][2]:
    x: int128[2] = [1,2]
    return [x, [3,4]]

@external
def noo(inp: int128[2]) -> int128[2]:
    return inp

@external
def poo(inp: int128[2][2]) -> int128[2][2]:
    return inp

@external
def qoo(inp: int128[2]) -> int128[2][2]:
    return [inp, [3,4]]

@external
def roo(inp: decimal[2]) -> decimal[2][2]:
    return [inp, [3.0, 4.0]]
    """

    c = get_contract(list_output_tester_code)
    assert c.foo() == [3, 5]
    assert c.goo() == [3, 5]
    assert c.hoo() == [3, 5]
    assert c.joo() == [3, 5]
    assert c.koo() == [[1, 2], [3, 4]]
    assert c.loo() == [[1, 2], [3, 4]]
    assert c.moo() == [[1, 2], [3, 4]]
    assert c.noo([3, 5]) == [3, 5]
    assert c.poo([[1, 2], [3, 4]]) == [[1, 2], [3, 4]]
    assert c.qoo([1, 2]) == [[1, 2], [3, 4]]
    assert c.roo(_map_nested(decimal_to_int, [1.0, 2.0])) == _map_nested(
        decimal_to_int, [[1.0, 2.0], [3.0, 4.0]]
    )


def test_array_accessor(get_contract):
    array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[4] = [0, 0, 0, 0]
    a[0] = x
    a[1] = y
    a[2] = z
    a[3] = w
    return a[0] * 1000 + a[1] * 100 + a[2] * 10 + a[3]
    """

    c = get_contract(array_accessor)
    assert c.test_array(2, 7, 1, 8) == 2718
    print("Passed basic array accessor test")


def test_two_d_array_accessor(get_contract):
    two_d_array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[2][2] = [[0, 0], [0, 0]]
    a[0][0] = x
    a[0][1] = y
    a[1][0] = z
    a[1][1] = w
    return a[0][0] * 1000 + a[0][1] * 100 + a[1][0] * 10 + a[1][1]
    """

    c = get_contract(two_d_array_accessor)
    assert c.test_array(2, 7, 1, 8) == 2718
    print("Passed complex array accessor test")


def test_three_d_array_accessor(get_contract):
    three_d_array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[2][2][2] = [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]
    a[0][0][0] = x
    a[0][0][1] = y
    a[0][1][0] = z
    a[0][1][1] = w
    a[1][0][0] = -x
    a[1][0][1] = -y
    a[1][1][0] = -z
    a[1][1][1] = -w
    return a[0][0][0] * 1000 + a[0][0][1] * 100 + a[0][1][0] * 10 + a[0][1][1] + \\
        a[1][1][1] * 1000 + a[1][1][0] * 100 + a[1][0][1] * 10 + a[1][0][0]
    """

    c = get_contract(three_d_array_accessor)
    assert c.test_array(2, 7, 1, 8) == -5454


def test_four_d_array_accessor(get_contract):
    four_d_array_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[2][2][2][2] = \\
        [[[[0, 0], [0, 0]], [[0, 0], [0, 0]]], [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]]
    a[0][0][0][0] = x
    a[0][0][0][1] = y
    a[0][0][1][0] = z
    a[0][0][1][1] = w
    a[0][1][0][0] = -x
    a[0][1][0][1] = -y
    a[0][1][1][0] = -z
    a[0][1][1][1] = -w

    a[1][0][0][0] = x + 1
    a[1][0][0][1] = y + 1
    a[1][0][1][0] = z + 1
    a[1][0][1][1] = w + 1
    a[1][1][0][0] = - (x + 1)
    a[1][1][0][1] = - (y + 1)
    a[1][1][1][0] = - (z + 1)
    a[1][1][1][1] = - (w + 1)
    return a[0][0][0][0] * 1000 + a[0][0][0][1] * 100 + a[0][0][1][0] * 10 + a[0][0][1][1] + \\
        a[0][1][1][1] * 1000 + a[0][1][1][0] * 100 + a[0][1][0][1] * 10 + a[0][1][0][0] + \\
        a[1][0][0][0] * 1000 + a[1][0][0][1] * 100 + a[1][0][1][0] * 10 + a[1][0][1][1] + \\
        a[1][1][1][1] * 1000 + a[1][1][1][0] * 100 + a[1][1][0][1] * 10 + a[1][1][0][0]
    """

    c = get_contract(four_d_array_accessor)
    assert c.test_array(2, 7, 1, 8) == -10908


def test_array_negative_accessor(get_contract, assert_compile_failed):
    array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[4] = [0, 0, 0, 0]
    a[0] = x
    a[1] = y
    a[2] = z
    a[3] = w
    return a[-4] * 1000 + a[-3] * 100 + a[-2] * 10 + a[-1]
    """

    assert_compile_failed(lambda: get_contract(array_negative_accessor), ArrayIndexException)

    two_d_array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[2][2] = [[0, 0], [0, 0]]
    a[0][0] = x
    a[0][1] = y
    a[1][0] = z
    a[1][1] = w
    return a[-2][-2] * 1000 + a[-2][-1] * 100 + a[-1][-2] * 10 + a[-1][-1]
    """

    assert_compile_failed(lambda: get_contract(two_d_array_negative_accessor), ArrayIndexException)

    three_d_array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[2][2][2] = [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]
    a[0][0][0] = x
    a[0][0][1] = y
    a[0][1][0] = z
    a[0][1][1] = w
    a[1][0][0] = -x
    a[1][0][1] = -y
    a[1][1][0] = -z
    a[1][1][1] = -w
    return a[-2][-2][-2] * 1000 + a[-2][-2][-1] * 100 + a[-2][-1][-2] * 10 + a[-2][-1][-1] + \\
        a[-1][-1][-1] * 1000 + a[-1][-1][-2] * 100 + a[-1][-2][-1] * 10 + a[-1][-2][-2]
    """

    assert_compile_failed(
        lambda: get_contract(three_d_array_negative_accessor), ArrayIndexException
    )

    four_d_array_negative_accessor = """
@external
def test_array(x: int128, y: int128, z: int128, w: int128) -> int128:
    a: int128[2][2][2][2] = \\
        [[[[0, 0], [0, 0]], [[0, 0], [0, 0]]], [[[0, 0], [0, 0]], [[0, 0], [0, 0]]]]
    a[0][0][0][0] = x
    a[0][0][0][1] = y
    a[0][0][1][0] = z
    a[0][0][1][1] = w
    a[0][1][0][0] = -x
    a[0][1][0][1] = -y
    a[0][1][1][0] = -z
    a[0][1][1][1] = -w

    a[1][0][0][0] = x + 1
    a[1][0][0][1] = y + 1
    a[1][0][1][0] = z + 1
    a[1][0][1][1] = w + 1
    a[1][1][0][0] = - (x + 1)
    a[1][1][0][1] = - (y + 1)
    a[1][1][1][0] = - (z + 1)
    a[1][1][1][1] = - (w + 1)
    return a[-2][-2][-2][-2] * 1000 + a[-2][-2][-2][-1] * 100 + \\
        a[-2][-2][-1][-2] * 10 + a[-2][-2][-1][-1] + \\
        a[-2][-1][-1][-1] * 1000 + a[-2][-1][-1][-2] * 100 + \\
        a[-2][-1][-2][-1] * 10 + a[-2][-1][-2][-2] + \\
        a[-1][-2][-2][-2] * 1000 + a[-1][-2][-2][-1] * 100 + \\
        a[-1][-2][-1][-2] * 10 + a[-1][-2][-1][-1] + \\
        a[-1][-1][-1][-1] * 1000 + a[-1][-1][-1][-2] * 100 + \\
        a[-1][-1][-2][-1] * 10 + a[-1][-1][-2][-2]
    """

    assert_compile_failed(lambda: get_contract(four_d_array_negative_accessor), ArrayIndexException)


def test_returns_lists(get_contract):
    code = """
@external
def test_array_num_return() -> int128[2][2]:
    a: int128[2][2] = [[1, 2], [3, 4]]
    return a

@external
def test_array_decimal_return1() -> decimal[2][2]:
    a: decimal[2][2] = [[1.0, 2.0], [3.0, 4.0]]
    return a

@external
def test_array_decimal_return2() -> decimal[2][2]:
    return [[1.0, 2.0], [3.0, 4.0]]

@external
def test_array_decimal_return3() -> decimal[2][2]:
    a: decimal[2][2] = [[1.0, 2.0], [3.0, 4.0]]
    return a
"""

    c = get_contract(code)
    assert c.test_array_num_return() == [[1, 2], [3, 4]]
    assert c.test_array_decimal_return1() == _map_nested(decimal_to_int, [[1.0, 2.0], [3.0, 4.0]])
    assert c.test_array_decimal_return2() == _map_nested(decimal_to_int, [[1.0, 2.0], [3.0, 4.0]])
    assert c.test_array_decimal_return3() == _map_nested(decimal_to_int, [[1.0, 2.0], [3.0, 4.0]])


def test_mult_list(get_contract):
    code = """
@external
def test_multi3() -> uint256[2][2][2]:
    l: uint256[2][2][2] = [[[0, 0], [0, 4]], [[0, 0], [0, 123]]]
    return l

@external
def test_multi4() -> uint256[2][2][2][2]:
    l: uint256[2][2][2][2] = [[[[1, 0], [0, 4]], [[0, 0], [0, 0]]], [[[444, 0], [0, 0]],[[1, 0], [0, 222]]]]  # noqa: E501
    return l
    """

    c = get_contract(code)

    assert c.test_multi3() == [[[0, 0], [0, 4]], [[0, 0], [0, 123]]]
    assert c.test_multi4() == [
        [[[1, 0], [0, 4]], [[0, 0], [0, 0]]],
        [[[444, 0], [0, 0]], [[1, 0], [0, 222]]],
    ]


@pytest.mark.parametrize("type_", ["uint8", "uint256"])
def test_unsigned_accessors(get_contract, tx_failed, type_):
    code = f"""
@external
def bounds_check(ix: {type_}) -> uint256:
    xs: uint256[3] = [1,2,3]
    return xs[ix]
    """
    c = get_contract(code)
    assert c.bounds_check(0) == 1
    assert c.bounds_check(2) == 3
    with tx_failed():
        c.bounds_check(3)


@pytest.mark.parametrize("type_", ["int128", "int256"])
def test_signed_accessors(get_contract, tx_failed, type_):
    code = f"""
@external
def bounds_check(ix: {type_}) -> uint256:
    xs: uint256[3] = [1,2,3]
    return xs[ix]
    """
    c = get_contract(code)
    assert c.bounds_check(0) == 1
    assert c.bounds_check(2) == 3
    with tx_failed():
        c.bounds_check(3)
    with tx_failed():
        c.bounds_check(-1)


def test_list_check_heterogeneous_types(get_contract, assert_compile_failed):
    code = """
@external
def fail() -> uint256:
    xs: uint256[3] = [1,2,3]
    return xs[3]
    """
    assert_compile_failed(lambda: get_contract(code), ArrayIndexException)
    code = """
@external
def fail() -> uint256:
    xs: uint256[3] = [1,2,3]
    return xs[-1]
    """
    assert_compile_failed(lambda: get_contract(code), ArrayIndexException)


def test_compile_time_bounds_check(get_contract, assert_compile_failed):
    code = """
@external
def parse_list_fail():
    xs: uint256[3] = [2**256, 1, 3]
    pass
    """
    assert_compile_failed(lambda: get_contract(code), OverflowException)


def test_2d_array_input_1(get_contract):
    code = """
@internal
def test_input(arr: int128[2][1], i: int128) -> (int128[2][1], int128):
    return arr, i

@external
def test_values(arr: int128[2][1], i: int128) -> (int128[2][1], int128):
    return self.test_input(arr, i)
    """

    c = get_contract(code)
    assert c.test_values([[1, 2]], 3) == ([[1, 2]], 3)


def test_2d_array_input_2(get_contract):
    code = """
@internal
def test_input(arr: int128[2][3], s: String[10]) -> (int128[2][3], String[10]):
    return arr, s

@external
def test_values(arr: int128[2][3], s: String[10]) -> (int128[2][3], String[10]):
    return self.test_input(arr, s)
    """

    c = get_contract(code)
    assert c.test_values([[1, 2], [3, 4], [5, 6]], "abcdef") == ([[1, 2], [3, 4], [5, 6]], "abcdef")


def test_nested_index_of_returned_array(get_contract):
    code = """
@internal
def inner() -> (int128, int128):
    return 1,2

@external
def outer() -> int128[2]:
    return [333, self.inner()[0]]
    """

    c = get_contract(code)
    assert c.outer() == [333, 1]


def test_nested_calls_inside_arrays(get_contract):
    code = """
@internal
def _foo(a: uint256, b: uint256[2]) -> (uint256, uint256, uint256, uint256, uint256):
    return 1, a, b[0], b[1], 5

@internal
def _foo2() -> uint256:
    a: uint256[10] = [6,7,8,9,10,11,12,13,15,16]
    return 4

@external
def foo() -> (uint256, uint256, uint256, uint256, uint256):
    return self._foo(2, [3, self._foo2()])
    """

    c = get_contract(code)
    assert c.foo() == (1, 2, 3, 4, 5)


def test_nested_calls_inside_arrays_with_index_access(get_contract):
    code = """
@internal
def _foo(a: uint256[2], b: uint256[2]) -> (uint256, uint256, uint256, uint256, uint256):
    return a[1]-b[0], 2, a[0]-b[1], 8-b[1], 5

@internal
def _foo2() -> (uint256, uint256):
    a: uint256[10] = [6,7,8,9,10,11,12,13,15,16]
    return a[6], 4

@external
def foo() -> (uint256, uint256, uint256, uint256, uint256):
    return self._foo([7, self._foo2()[0]], [11, self._foo2()[1]])
    """

    c = get_contract(code)
    assert c.foo() == (1, 2, 3, 4, 5)


def test_so_many_things_you_should_never_do(get_contract):
    code = """
@internal
def _foo(a: uint256[2], b: uint256[2]) -> uint256[5]:
    return [a[1]-b[0], 2, a[0]-b[1], 8-b[1], 5]

@internal
def _foo2() -> (uint256, uint256):
    b: uint256[2] = [5, 8]
    a: uint256[10] = [6,7,8,9,10,11,12,13,self._foo([44,b[0]],b)[4],16]
    return a[6], 4

@external
def foo() -> (uint256, uint256[3], uint256[2]):
    x: uint256[3] = [1, 14-self._foo2()[0], self._foo([7,self._foo2()[0]], [11,self._foo2()[1]])[2]]
    return 666, x, [88, self._foo2()[0]]
    """
    c = get_contract(code)
    assert c.foo() == (666, [1, 2, 3], [88, 12])


def test_list_of_dynarray(get_contract):
    code = """
@external
def bar(x: int128) -> DynArray[int128, 2][2]:
    a: DynArray[int128, 2][2] = [[x, x * 2], [x * 3, x * 4]]
    return a

@external
def foo(x: int128) -> int128:
    a: DynArray[int128, 2][2] = [[x, x * 2], [x * 3, x * 4]]
    return a[0][0] * a[1][1]
    """
    c = get_contract(code)
    assert c.bar(7) == [[7, 14], [21, 28]]
    assert c.foo(7) == 196


def test_list_of_nested_dynarray(get_contract):
    code = """
@external
def bar(x: int128) -> DynArray[int128, 2][2][2]:
    a: DynArray[int128, 2][2][2] = [
        [[x, x * 2], [x * 3, x * 4]],
        [[x * 5, x * 6], [x * 7, x * 8]],
    ]
    return a

@external
def foo(x: int128) -> int128:
    a: DynArray[int128, 2][2][2] = [
        [[x, x * 2], [x * 3, x * 4]],
        [[x * 5, x * 6], [x * 7, x * 8]],
    ]
    return a[0][0][0] * a[1][1][1]
    """
    c = get_contract(code)
    assert c.bar(7) == [[[7, 14], [21, 28]], [[35, 42], [49, 56]]]
    assert c.foo(7) == 392


def test_list_of_structs_arg(get_contract):
    code = """
struct Foo:
    x: uint256
    y: uint256

@external
def bar(_baz: Foo[3]) -> uint256:
    sum: uint256 = 0
    for i: uint256 in range(3):
        sum += _baz[i].x * _baz[i].y
    return sum
    """
    c = get_contract(code)
    c_input = [[x, y] for x, y in zip(range(3), range(3))]
    assert c.bar(c_input) == 5  # 0 * 0 + 1 * 1 + 2 * 2


def test_list_of_structs_arg_with_dynamic_type(get_contract):
    code = """
struct Foo:
    x: uint256
    _msg: String[32]

@external
def bar(_baz: Foo[3]) -> String[96]:
    return concat(_baz[0]._msg, _baz[1]._msg, _baz[2]._msg)
    """
    c = get_contract(code)
    c_input = [[i, msg] for i, msg in enumerate(("Hello ", "world", "!!!!"))]
    assert c.bar(c_input) == "Hello world!!!!"


def test_list_of_nested_struct_arrays(get_contract):
    code = """
struct Ded:
    a: uint256[3]
    b: bool

struct Foo:
    c: uint256
    d: uint256
    e: Ded

struct Bar:
    f: Foo[3]
    g: DynArray[uint256, 3]

@external
def bar(_bar: Bar[3]) -> uint256:
    sum: uint256 = 0
    for i: uint256 in range(3):
        sum += _bar[i].f[0].e.a[0] * _bar[i].f[1].e.a[1]
    return sum
    """
    c = get_contract(code)
    c_input = [
        ((tuple([(123, 456, ([i, i + 1, i + 2], False))] * 3)), [9, 8, 7]) for i in range(1, 4)
    ]

    assert c.bar(c_input) == 20


def test_2d_list_of_struct(get_contract):
    code = """
struct Bar:
    a: uint256
    b: uint256

@external
def foo(x: Bar[2][2]) -> uint256:
    return x[0][0].a + x[1][1].b
    """
    c = get_contract(code)
    c_input = [([i, i * 2], [i * 3, i * 4]) for i in range(1, 3)]
    assert c.foo(c_input) == 9


def test_3d_list_of_struct(get_contract):
    code = """
struct Bar:
    a: uint256
    b: uint256

@external
def foo(x: Bar[2][2][2]) -> uint256:
    return x[0][0][0].a + x[1][1][1].b
    """
    c = get_contract(code)
    c_input = [([([i, i * 2], [i * 3, i * 4]) for i in range(1, 3)])] * 2
    assert c.foo(c_input) == 9


@pytest.mark.parametrize(
    "type,value",
    [
        ("decimal", [5.0, 11.0, 17.0, 29.0, 37.0, 41.0]),
        ("uint8", [0, 1, 17, 250, 255, 2]),
        ("int128", [0, -1, 1, -(2**127), 2**127 - 1, -50]),
        ("int256", [0, -1, 1, -(2**255), 2**255 - 1, -50]),
        ("uint256", [0, 1, 2**8, 2**255 + 1, 2**256 - 1, 100]),
        (
            "uint256",
            [2**255 + 1, 2**255 + 2, 2**255 + 3, 2**255 + 4, 2**255 + 5, 2**255 + 6],
        ),
        ("bool", [True, False, True, False, True, False]),
    ],
)
def test_constant_list(get_contract, tx_failed, type, value):
    code = f"""
MY_LIST: constant({type}[{len(value)}]) = {value}
@external
def ix(i: uint256) -> {type}:
    return MY_LIST[i]
    """
    c = get_contract(code)
    for i, p in enumerate(value):
        if type == "decimal":
            # special case to transform for decimal ABI
            p = decimal_to_int(p)
        assert c.ix(i) == p
    # assert oob
    with tx_failed():
        c.ix(len(value) + 1)


def test_nested_constant_list_accessor(get_contract):
    code = """
@external
def foo() -> bool:
    f: uint256 = 1
    a: bool = 1 == [1,2,4][f] + -1
    return a
    """
    c = get_contract(code)
    assert c.foo() is True


# Would be nice to put this somewhere accessible, like in vyper.types or something
integer_types = ["uint8", "int128", "int256", "uint256"]


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_list_fail(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant({storage_type}[3]) = [1, 2, 3]

@external
def foo() -> {return_type}[3]:
    return MY_CONSTANT
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_list_fail_2(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant({storage_type}[3]) = [1, 2, 3]

@external
def foo() -> {return_type}:
    return MY_CONSTANT[0]
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_list_fail_3(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant({storage_type}[3]) = [1, 2, 3]

@external
def foo(i: uint256) -> {return_type}:
    return MY_CONSTANT[i]
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


def test_constant_list_address(get_contract, tx_failed):
    some_good_address = [
        "0x0000000000000000000000000000000000012345",
        "0x0000000000000000000000000000000000023456",
        "0x0000000000000000000000000000000000034567",
        "0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF",
        "0xffffFFFfFFffffffffffffffFfFFFfffFFFfFFfE",
        "0xFfffFfFFFfFFFFfFFfFFFfFFFfFFfFFFfFfFfFf1",
    ]
    code = """
MY_LIST: constant(address[6]) = [
    0x0000000000000000000000000000000000012345,
    0x0000000000000000000000000000000000023456,
    0x0000000000000000000000000000000000034567,
    0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF,
    0xffffFFFfFFffffffffffffffFfFFFfffFFFfFFfE,
    0xFfffFfFFFfFFFFfFFfFFFfFFFfFFfFFFfFfFfFf1
]
@external
def ix(i: uint256) -> address:
    return MY_LIST[i]
    """
    c = get_contract(code)
    for i, p in enumerate(some_good_address):
        assert c.ix(i) == p
    # assert oob
    with tx_failed():
        c.ix(len(some_good_address) + 1)


def test_list_index_complex_expr(get_contract, tx_failed):
    # test subscripts where the index is not a literal
    code = """
@external
def foo(xs: uint256[257], i: uint8) -> uint256:
    return xs[i + 1]
    """
    c = get_contract(code)
    xs = [i + 1 for i in range(257)]

    for ix in range(255):
        assert c.foo(xs, ix) == xs[ix + 1]

    # safemath should fail for uint8: 255 + 1.
    with tx_failed():
        c.foo(xs, 255)


@pytest.mark.parametrize(
    "type,value",
    [
        ("decimal", [[5.0, 11.0], [17.0, 29.0], [37.0, 41.0]]),
        ("uint8", [[0, 1], [17, 250], [255, 2]]),
        ("int128", [[0, -1], [1, -(2**127)], [2**127 - 1, -50]]),
        ("int256", [[0, -1], [1, -(2**255)], [2**255 - 1, -50]]),
        ("uint256", [[0, 1], [2**8, 2**255 + 1], [2**256 - 1, 100]]),
        (
            "uint256",
            [
                [2**255 + 1, 2**255 + 2],
                [2**255 + 3, 2**255 + 4],
                [2**255 + 5, 2**255 + 6],
            ],
        ),
        ("bool", [[True, False], [True, False], [True, False]]),
    ],
)
def test_constant_nested_list(get_contract, tx_failed, type, value):
    code = f"""
MY_LIST: constant({type}[{len(value[0])}][{len(value)}]) = {value}
@external
def ix(i: uint256, j: uint256) -> {type}:
    return MY_LIST[i][j]
    """
    c = get_contract(code)
    for i, p in enumerate(value):
        for j, q in enumerate(p):
            if type == "decimal":
                # special case for decimal ABI
                q = decimal_to_int(q)
            assert c.ix(i, j) == q
    # assert oob
    with tx_failed():
        c.ix(len(value) + 1, len(value[0]) + 1)


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_nested_list_fail(get_contract, assert_compile_failed, storage_type, return_type):
    code = f"""
MY_CONSTANT: constant({storage_type}[2][3]) = [[1, 2], [3, 4], [5, 6]]

@external
def foo() -> {return_type}[2][3]:
    return MY_CONSTANT
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


@pytest.mark.parametrize("storage_type,return_type", itertools.permutations(integer_types, 2))
def test_constant_nested_list_fail_2(
    get_contract, assert_compile_failed, storage_type, return_type
):
    code = f"""
MY_CONSTANT: constant({storage_type}[2][3]) = [[1, 2], [3, 4], [5, 6]]

@external
def foo() -> {return_type}:
    return MY_CONSTANT[0][0]
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)



import pytest

from vyper.ast import parse_to_ast
from vyper.exceptions import CallViolation, StructureException
from vyper.semantics.analysis import analyze_module


def test_self_function_call(dummy_input_bundle):
    code = """
@internal
def foo():
    self.foo()
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(CallViolation):
        analyze_module(vyper_module, dummy_input_bundle)


def test_cyclic_function_call(dummy_input_bundle):
    code = """
@internal
def foo():
    self.bar()

@internal
def bar():
    self.foo()
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(CallViolation):
        analyze_module(vyper_module, dummy_input_bundle)


def test_multi_cyclic_function_call(dummy_input_bundle):
    code = """
@internal
def foo():
    self.bar()

@internal
def bar():
    self.baz()

@internal
def baz():
    self.potato()

@internal
def potato():
    self.foo()
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(CallViolation):
        analyze_module(vyper_module, dummy_input_bundle)


def test_global_ann_assign_callable_no_crash(dummy_input_bundle):
    code = """
balanceOf: public(HashMap[address, uint256])

@internal
def foo(to : address):
    self.balanceOf(to)
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(StructureException) as excinfo:
        analyze_module(vyper_module, dummy_input_bundle)
    assert excinfo.value.message == "HashMap[address, uint256] is not callable"



from vyper.exceptions import CompilerPanic
from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.basicblock import IRBasicBlock, IRLabel
from vyper.venom.passes.base_pass import IRPass


class NormalizationPass(IRPass):
    """
    This pass splits basic blocks when there are multiple conditional predecessors.
    The code generator expect a normalized CFG, that has the property that
    each basic block has at most one conditional predecessor.
    """

    changes = 0

    def _split_basic_block(self, bb: IRBasicBlock) -> None:
        # Iterate over the predecessors to this basic block
        for in_bb in list(bb.cfg_in):
            assert bb in in_bb.cfg_out
            # Handle branching in the predecessor bb
            if len(in_bb.cfg_out) > 1:
                self._insert_split_basicblock(bb, in_bb)
                self.changes += 1
                break

    def _insert_split_basicblock(self, bb: IRBasicBlock, in_bb: IRBasicBlock) -> IRBasicBlock:
        # Create an intermediary basic block and append it
        source = in_bb.label.value
        target = bb.label.value
        fn = self.function

        split_label = IRLabel(f"{source}_split_{target}")
        in_terminal = in_bb.instructions[-1]
        in_terminal.replace_label_operands({bb.label: split_label})

        split_bb = IRBasicBlock(split_label, fn)
        split_bb.append_instruction("jmp", bb.label)
        fn.append_basic_block(split_bb)

        for inst in bb.instructions:
            if inst.opcode != "phi":
                continue
            for i in range(0, len(inst.operands), 2):
                if inst.operands[i] == in_bb.label:
                    inst.operands[i] = split_bb.label

        # Update the labels in the data segment
        for inst in fn.ctx.data_segment:
            if inst.opcode == "db" and inst.operands[0] == bb.label:
                inst.operands[0] = split_bb.label

        return split_bb

    def _run_pass(self) -> int:
        fn = self.function
        self.changes = 0

        self.analyses_cache.request_analysis(CFGAnalysis)

        # Split blocks that need splitting
        for bb in list(fn.get_basic_blocks()):
            if len(bb.cfg_in) > 1:
                self._split_basic_block(bb)

        # If we made changes, recalculate the cfg
        if self.changes > 0:
            self.analyses_cache.force_analysis(CFGAnalysis)
            fn.remove_unreachable_blocks()

        return self.changes

    def run_pass(self):
        fn = self.function
        for _ in range(fn.num_basic_blocks * 2):
            if self._run_pass() == 0:
                break
        else:
            raise CompilerPanic("Normalization pass did not converge")



from eth_utils import to_wei


def test_throw_on_sending(env, tx_failed, get_contract):
    code = """
x: public(int128)

@deploy
def __init__():
    self.x = 123
    """
    c = get_contract(code)

    assert c.x() == 123
    assert env.get_balance(c.address) == 0
    value = to_wei(0.1, "ether")
    env.set_balance(env.deployer, value)
    with tx_failed():
        env.message_call(c.address, value=value, data=b"")  # call default function
    assert env.get_balance(c.address) == 0


def test_basic_default(env, get_logs, get_contract):
    code = """
event Sent:
    sender: indexed(address)

@external
@payable
def __default__():
    log Sent(msg.sender)
    """
    c = get_contract(code)
    env.set_balance(env.deployer, 10**18)
    env.message_call(c.address, value=10**17, data=b"")  # call default function
    (log,) = get_logs(c, "Sent")
    assert env.deployer == log.args.sender
    assert env.get_balance(c.address) == to_wei(0.1, "ether")


def test_basic_default_default_param_function(env, get_logs, get_contract):
    code = """
event Sent:
    sender: indexed(address)

@external
@payable
def fooBar(a: int128 = 12345) -> int128:
    log Sent(empty(address))
    return a

@external
@payable
def __default__():
    log Sent(msg.sender)
    """
    c = get_contract(code)
    env.set_balance(env.deployer, 10**18)
    env.message_call(c.address, value=10**17, data=b"")  # call default function
    (log,) = get_logs(c, "Sent")
    assert env.deployer == log.args.sender
    assert env.get_balance(c.address) == to_wei(0.1, "ether")


def test_basic_default_not_payable(env, tx_failed, get_contract):
    code = """
event Sent:
    sender: indexed(address)

@external
def __default__():
    log Sent(msg.sender)
    """
    c = get_contract(code)
    env.set_balance(env.deployer, 10**17)

    with tx_failed():
        env.message_call(c.address, value=10**17, data=b"")  # call default function


def test_multi_arg_default(assert_compile_failed, get_contract):
    code = """
@payable
@external
def __default__(arg1: int128):
    pass
    """
    assert_compile_failed(lambda: get_contract(code))


def test_always_public(assert_compile_failed, get_contract):
    code = """
@internal
def __default__():
    pass
    """
    assert_compile_failed(lambda: get_contract(code))


def test_always_public_2(assert_compile_failed, get_contract):
    code = """
event Sent:
    sender: indexed(address)

def __default__():
    log Sent(msg.sender)
    """
    assert_compile_failed(lambda: get_contract(code))


def test_zero_method_id(env, get_logs, get_contract, tx_failed):
    # test a method with 0x00000000 selector,
    # expects at least 36 bytes of calldata.
    code = """
event Sent:
    sig: uint256

@external
@payable
# function selector: 0x00000000
def blockHashAskewLimitary(v: uint256) -> uint256:
    log Sent(2)
    return 7

@external
def __default__():
    log Sent(1)
    """
    c = get_contract(code)

    assert c.blockHashAskewLimitary(0) == 7

    def _call_with_bytes(hexstr):
        # call our special contract and return the logged value
        data = bytes.fromhex(hexstr.removeprefix("0x"))
        env.message_call(c.address, value=0, data=data)
        (log,) = get_logs(c, "Sent")
        return log.args.sig

    assert 1 == _call_with_bytes("0x")

    # call blockHashAskewLimitary with proper calldata
    assert 2 == _call_with_bytes("0x" + "00" * 36)

    # call blockHashAskewLimitary with extra trailing bytes in calldata
    assert 2 == _call_with_bytes("0x" + "00" * 37)

    for i in range(4):
        # less than 4 bytes of calldata doesn't match the 0 selector and goes to default
        assert 1 == _call_with_bytes("0x" + "00" * i)

    for i in range(4, 36):
        # match the full 4 selector bytes, but revert due to malformed (short) calldata
        with tx_failed():
            _call_with_bytes(f"0x{'00' * i}")


def test_another_zero_method_id(env, get_logs, get_contract, tx_failed):
    # test another zero method id but which only expects 4 bytes of calldata
    code = """
event Sent:
    sig: uint256

@external
@payable
# function selector: 0x00000000
def wycpnbqcyf() -> uint256:
    log Sent(2)
    return 7

@external
def __default__():
    log Sent(1)
    """
    c = get_contract(code)

    assert c.wycpnbqcyf() == 7

    def _call_with_bytes(hexstr):
        # call our special contract and return the logged value
        data = bytes.fromhex(hexstr.removeprefix("0x"))
        env.message_call(c.address, value=0, data=data, gas=10**6)
        (log,) = get_logs(c, "Sent")
        return log.args.sig

    assert 1 == _call_with_bytes("0x")

    # call wycpnbqcyf
    assert 2 == _call_with_bytes("0x" + "00" * 4)

    # too many bytes ok
    assert 2 == _call_with_bytes("0x" + "00" * 5)

    # "right" method id but by accident - not enough bytes.
    for i in range(4):
        assert 1 == _call_with_bytes("0x" + "00" * i)


def test_partial_selector_match_trailing_zeroes(env, get_logs, get_contract):
    code = """
event Sent:
    sig: uint256

@external
@payable
# function selector: 0xd88e0b00
def fow() -> uint256:
    log Sent(2)
    return 7

@external
def __default__():
    log Sent(1)
    """
    c = get_contract(code)

    # sanity check - we can call c.fow()
    assert c.fow() == 7

    def _call_with_bytes(hexstr):
        # call our special contract and return the logged value
        data = bytes.fromhex(hexstr.removeprefix("0x"))
        env.message_call(c.address, value=0, data=data)
        (log,) = get_logs(c, "Sent")
        return log.args.sig

    # check we can call default function
    assert 1 == _call_with_bytes("0x")

    # check fow() selector is 0xd88e0b00
    assert 2 == _call_with_bytes("0xd88e0b00")

    # check calling d88e0b with no trailing zero goes to fallback instead of reverting
    assert 1 == _call_with_bytes("0xd88e0b")



from typing import Dict, Optional, Sequence, Tuple, Union

# Parser
ModificationOffsets = Dict[Tuple[int, int], str]
ParserPosition = Tuple[int, int]

# Compiler
ContractPath = str
SourceCode = str
OutputFormats = Sequence[str]
StorageLayout = Dict

# Opcodes
OpcodeGasCost = Union[int, Tuple]
OpcodeValue = Tuple[Optional[int], int, int, OpcodeGasCost]
OpcodeMap = Dict[str, OpcodeValue]
OpcodeRulesetValue = Tuple[Optional[int], int, int, int]
OpcodeRulesetMap = Dict[str, OpcodeRulesetValue]



import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold

variables = "abcdefghij"


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(values=st.lists(st.booleans(), min_size=2, max_size=10))
@pytest.mark.parametrize("comparator", ["and", "or"])
def test_boolop_simple(get_contract, values, comparator):
    input_value = ",".join(f"{i}: bool" for i in variables[: len(values)])
    return_value = f" {comparator} ".join(variables[: len(values)])

    source = f"""
@external
def foo({input_value}) -> bool:
    return {return_value}
    """
    contract = get_contract(source)

    literal_op = f" {comparator} ".join(str(i) for i in values)

    vyper_ast = parse_and_fold(literal_op)
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(*values) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(
    values=st.lists(st.booleans(), min_size=2, max_size=10),
    comparators=st.lists(st.sampled_from(["and", "or"]), min_size=11, max_size=11),
)
def test_boolop_nested(get_contract, values, comparators):
    input_value = ",".join(f"{i}: bool" for i in variables[: len(values)])
    return_value = " ".join(f"{a} {b}" for a, b in zip(variables[: len(values)], comparators))
    return_value = return_value.rsplit(maxsplit=1)[0]

    source = f"""
@external
def foo({input_value}) -> bool:
    return {return_value}
    """
    contract = get_contract(source)

    literal_op = " ".join(f"{a} {b}" for a, b in zip(values, comparators))
    literal_op = literal_op.rsplit(maxsplit=1)[0]

    vyper_ast = parse_and_fold(literal_op)
    new_node = vyper_ast.body[0].value.get_folded_value()
    expected = new_node.value

    assert contract.foo(*values) == expected



import pytest

from vyper import compiler
from vyper.exceptions import StructureException

FAILING_CONTRACTS = [
    """
@external
@view
def gsf():
    pass

@external
@view
def tgeo():
    pass
    """,
    """
@external
@view
def withdraw(a: uint256):
    pass

@external
@view
def OwnerTransferV7b711143(a: uint256):
    pass
    """,
    """
@external
@view
def withdraw(a: uint256):
    pass

@external
@view
def gsf():
    pass

@external
@view
def tgeo():
    pass

@external
@view
def OwnerTransferV7b711143(a: uint256):
    pass
    """,
    """
# check collision with ID = 0x00000000
wycpnbqcyf:public(uint256)

@external
def randallsRevenge_ilxaotc(): pass
    """,
]


@pytest.mark.parametrize("failing_contract_code", FAILING_CONTRACTS)
def test_method_id_conflicts(failing_contract_code):
    with pytest.raises(StructureException) as e:
        compiler.compile_code(failing_contract_code)

    assert e.value._message.startswith("Methods produce colliding method ID")


export_fails = [
    {
        "lib1.vy": """
wycpnbqcyf:public(uint256)
        """,
        "MAIN": """
import lib1

initializes: lib1
exports: lib1.wycpnbqcyf

@external
def randallsRevenge_ilxaotc(): pass
    """,
    },
    {
        "lib1.vy": """
@external
@view
def withdraw(a: uint256):
    pass

@external
@view
def gsf():
    pass
    """,
        "MAIN": """
import lib1

exports: (lib1.gsf, lib1.withdraw)

@external
@view
def tgeo():
    pass

@external
@view
def OwnerTransferV7b711143(a: uint256):
    pass
    """,
    },
]


@pytest.mark.parametrize("files", export_fails)
def test_method_id_conflicts_export(files, make_input_bundle):
    main = files.pop("MAIN")
    input_bundle = make_input_bundle(files)
    with pytest.raises(StructureException) as e:
        compiler.compile_code(main, input_bundle=input_bundle)

    assert e.value._message.startswith("Methods produce colliding method ID")



def test_gas_call(get_contract):
    gas_call = """
@external
def foo() -> uint256:
    return msg.gas
    """

    c = get_contract(gas_call)

    assert c.foo(gas=50000) < 50000
    assert c.foo(gas=50000) > 25000



import contextlib
from dataclasses import dataclass, field
from typing import Iterator

from vyper import ast as vy_ast
from vyper.exceptions import CompilerPanic, ImportCycle

"""
data structure for collecting import statements and validating the
import graph
"""


@dataclass
class ImportGraph:
    # the current path in the import graph traversal
    _path: list[vy_ast.Module] = field(default_factory=list)

    def push_path(self, module_ast: vy_ast.Module) -> None:
        if module_ast in self._path:
            cycle = self._path + [module_ast]
            raise ImportCycle(" imports ".join(f'"{t.path}"' for t in cycle))

        self._path.append(module_ast)

    def pop_path(self, expected: vy_ast.Module) -> None:
        popped = self._path.pop()
        if expected != popped:
            raise CompilerPanic("unreachable")

    @contextlib.contextmanager
    def enter_path(self, module_ast: vy_ast.Module) -> Iterator[None]:
        self.push_path(module_ast)
        try:
            yield
        finally:
            self.pop_path(module_ast)



# test syntactic comparisons
# most tests under tests/ast/nodes/test_evaluate_compare.py
import pytest


def test_3034_verbatim(get_contract):
    # test GH issue 3034 exactly
    code = """
@view
@external
def showError():
    adr1: address = 0xFbEEa1C75E4c4465CB2FCCc9c6d6afe984558E20
    adr2: address = 0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2
    adr3: address = 0xFbEEa1C75E4c4465CB2FCCc9c6d6afe984558E20
    assert adr1 in [adr2,adr3], "error in comparison with in statement!"
    """
    c = get_contract(code)
    c.showError()


@pytest.mark.parametrize("invert", (True, False))
def test_in_list(get_contract, invert):
    # test slightly more complicated variations of #3034
    INVERT = "not" if invert else ""
    code = f"""
SOME_ADDRESS: constant(address) = 0x22cb70ba2EC32347D9e32740fc14b2f3d038Ce8E
@view
@external
def test_in(addr: address) -> bool:
    x: address = 0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2
    y: address = 0xFbEEa1C75E4c4465CB2FCCc9c6d6afe984558E20
    # in list which
    return addr {INVERT} in [x, y, SOME_ADDRESS]
    """
    c = get_contract(code)
    should_in = [
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
        "0xFbEEa1C75E4c4465CB2FCCc9c6d6afe984558E20",
        "0x22cb70ba2EC32347D9e32740fc14b2f3d038Ce8E",
    ]
    should_not_in = [
        "0x" + "00" * 20,
        "0xfBeeA1C75E4C4465CB2fccC9C6d6AFe984558e21",  # y but last bit flipped
    ]
    for t in should_in:
        assert c.test_in(t) is (True if not invert else False)
    for t in should_not_in:
        assert c.test_in(t) is (True if invert else False)



from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.basicblock import IRBasicBlock, IRLabel
from vyper.venom.context import IRContext
from vyper.venom.passes.make_ssa import MakeSSA


def test_phi_case():
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb = fn.get_basic_block()

    bb_cont = IRBasicBlock(IRLabel("condition"), fn)
    bb_then = IRBasicBlock(IRLabel("then"), fn)
    bb_else = IRBasicBlock(IRLabel("else"), fn)
    bb_if_exit = IRBasicBlock(IRLabel("if_exit"), fn)
    fn.append_basic_block(bb_cont)
    fn.append_basic_block(bb_then)
    fn.append_basic_block(bb_else)
    fn.append_basic_block(bb_if_exit)

    v = bb.append_instruction("mload", 64)
    bb_cont.append_instruction("jnz", v, bb_then.label, bb_else.label)

    bb_if_exit.append_instruction("add", v, 1, ret=v)
    bb_if_exit.append_instruction("jmp", bb_cont.label)

    bb_then.append_instruction("assert", bb_then.append_instruction("mload", 96))
    bb_then.append_instruction("jmp", bb_if_exit.label)
    bb_else.append_instruction("jmp", bb_if_exit.label)

    bb.append_instruction("jmp", bb_cont.label)

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()

    condition_block = fn.get_basic_block("condition")
    assert len(condition_block.instructions) == 2

    phi_inst = condition_block.instructions[0]
    assert phi_inst.opcode == "phi"
    assert phi_inst.operands[0].name == "_global"
    assert phi_inst.operands[1].name == "%1"
    assert phi_inst.operands[2].name == "if_exit"
    assert phi_inst.operands[3].name == "%1"
    assert phi_inst.output.name == "%1"
    assert phi_inst.output.value != phi_inst.operands[1].value
    assert phi_inst.output.value != phi_inst.operands[3].value



import pytest

from vyper import compile_code
from vyper.exceptions import TypeMismatch

fail_list = [
    """
@external
def foo(inp: Bytes[4]) -> int128:
    return len(inp)
    """,
    """
@external
def foo(inp: int128) -> uint256:
    return len(inp)
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_block_fail(bad_code):
    if isinstance(bad_code, tuple):
        with pytest.raises(bad_code[1]):
            compile_code(bad_code[0])
    else:
        with pytest.raises(TypeMismatch):
            compile_code(bad_code)


valid_list = [
    """
@external
def foo(inp: Bytes[10]) -> uint256:
    return len(inp)
    """,
    """
@external
def foo(inp: String[10]) -> uint256:
    return len(inp)
    """,
    """
BAR: constant(String[5]) = "vyper"
FOO: constant(uint256) = len(BAR)

@external
def foo() -> uint256:
    a: uint256 = FOO
    return a
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_list_success(good_code):
    assert compile_code(good_code) is not None



import pytest
from eth.vm.forks.cancun.constants import BLOB_BASE_FEE_UPDATE_FRACTION, MIN_BLOB_BASE_FEE
from eth.vm.forks.cancun.state import fake_exponential


@pytest.mark.requires_evm_version("cancun")
def test_blobbasefee(env, get_contract):
    code = """
@external
@view
def get_blobbasefee() -> uint256:
    return block.blobbasefee
"""
    c = get_contract(code)

    assert c.get_blobbasefee() == MIN_BLOB_BASE_FEE

    env.set_balance(env.deployer, 10**20)
    env.set_excess_blob_gas(10**6)

    # kzg_hash(b"Vyper is the language of the sneks")
    env.blob_hashes = [
        (bytes.fromhex("015a5c97e3cc516f22a95faf7eefff00eb2fee7a65037fde07ac5446fc93f2a0"))
    ] * 6

    env.message_call(
        "0xb45BEc6eeCA2a09f4689Dd308F550Ad7855051B5",  # random address
        gas=21000,
        gas_price=10**10,
    )

    excess_blob_gas = env.get_excess_blob_gas()
    expected_blobbasefee = fake_exponential(
        MIN_BLOB_BASE_FEE, excess_blob_gas, BLOB_BASE_FEE_UPDATE_FRACTION
    )
    assert c.get_blobbasefee() == expected_blobbasefee



from eth_account import Account
from eth_account._utils.signing import to_bytes32

from tests.utils import ZERO_ADDRESS


def test_ecrecover_test(get_contract):
    ecrecover_test = """
@external
def test_ecrecover(h: bytes32, v: uint8, r: bytes32, s: bytes32) -> address:
    return ecrecover(h, v, r, s)

@external
def test_ecrecover_uints(h: bytes32, v: uint256, r: uint256, s: uint256) -> address:
    return ecrecover(h, v, r, s)

@external
def test_ecrecover2() -> address:
    return ecrecover(0x3535353535353535353535353535353535353535353535353535353535353535,
                     28,
                     0x8bb954e648c468c01b6efba6cd4951929d16e5235077e2be43e81c0c139dbcdf,
                     0x0e8a97aa06cc123b77ccf6c85b123d299f3f477200945ef71a1e1084461cba8d)

@external
def test_ecrecover_uints2() -> address:
    return ecrecover(0x3535353535353535353535353535353535353535353535353535353535353535,
                     28,
                     63198938615202175987747926399054383453528475999185923188997970550032613358815,
                     6577251522710269046055727877571505144084475024240851440410274049870970796685)

    """

    c = get_contract(ecrecover_test)

    h = b"\x35" * 32
    local_account = Account.from_key(b"\x46" * 32)
    sig = local_account.signHash(h)

    assert c.test_ecrecover(h, sig.v, to_bytes32(sig.r), to_bytes32(sig.s)) == local_account.address
    assert c.test_ecrecover_uints(h, sig.v, sig.r, sig.s) == local_account.address
    assert c.test_ecrecover2() == local_account.address
    assert c.test_ecrecover_uints2() == local_account.address

    print("Passed ecrecover test")


def test_invalid_signature(get_contract):
    code = """
dummies: HashMap[address, HashMap[address, uint256]]

@external
def test_ecrecover(hash: bytes32, v: uint8, r: uint256) -> address:
    # read from hashmap to put garbage in 0 memory location
    s: uint256 = self.dummies[msg.sender][msg.sender]
    return ecrecover(hash, v, r, s)
    """
    c = get_contract(code)
    hash_ = bytes(i for i in range(32))
    v = 0  # invalid v! ecrecover precompile will not write to output buffer
    r = 0
    assert c.test_ecrecover(hash_, v, r) == ZERO_ADDRESS


# slightly more subtle example: get_v() stomps memory location 0,
# so this tests that the output buffer stays clean during ecrecover()
# builtin execution.
def test_invalid_signature2(get_contract):
    code = """

owner: immutable(address)

@deploy
def __init__():
    owner = 0x7E5F4552091A69125d5DfCb7b8C2659029395Bdf

@internal
def get_v() -> uint256:
    assert owner == owner # force a dload to write at index 0 of memory
    return 21

@payable
@external
def test_ecrecover() -> bool:
    assert ecrecover(empty(bytes32), self.get_v(), 0, 0) == empty(address)
    return True
    """
    c = get_contract(code)
    assert c.test_ecrecover() is True



from pytest import raises

from vyper.exceptions import VyperException


def test_type_exception_pos():
    pos = (1, 2)

    with raises(VyperException) as e:
        raise VyperException("Fail!", pos)

    assert e.value.lineno == 1
    assert e.value.col_offset == 2
    assert str(e.value) == "line 1:2 Fail!"


# multiple exceptions in file
def test_multiple_exceptions(get_contract, assert_compile_failed):
    code = """
struct A:
    b: B  # unknown type

foo: immutable(uint256)
bar: immutable(uint256)
@deploy
def __init__():
    self.foo = 1  # SyntaxException
    self.bar = 2  # SyntaxException

    """
    assert_compile_failed(lambda: get_contract(code), VyperException)



from vyper.compiler import compile_code


def test_implements_from_vyi(make_input_bundle):
    vyi = """
@external
def foo():
    ...
    """
    lib1 = """
import some_interface
    """
    main = """
import lib1

implements: lib1.some_interface

@external
def foo():  # implementation
    pass
    """
    input_bundle = make_input_bundle({"some_interface.vyi": vyi, "lib1.vy": lib1})

    assert compile_code(main, input_bundle=input_bundle) is not None


def test_implements_from_vyi2(make_input_bundle):
    # test implements via nested imported vyi file
    vyi = """
@external
def foo():
    ...
    """
    lib1 = """
import some_interface
    """
    lib2 = """
import lib1
    """
    main = """
import lib2

implements: lib2.lib1.some_interface

@external
def foo():  # implementation
    pass
    """
    input_bundle = make_input_bundle({"some_interface.vyi": vyi, "lib1.vy": lib1, "lib2.vy": lib2})

    assert compile_code(main, input_bundle=input_bundle) is not None
